"""
Sopel - An IRC Bot
Copyright 2008, Sean B. Palmer, inamidst.com
Copyright © 2012-2014, Elad Alfassa <elad@fedoraproject.org>
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

import argparse
import logging
import os
import platform
import signal
import sys
import time

from sopel import __version__, bot, config, logger
from . import utils

# This is in case someone somehow manages to install Sopel on an old version
# of pip (<9.0.0), which doesn't know about `python_requires`, or tries to run
# from source on an unsupported version of Python.
if sys.version_info < (3, 8):
    utils.stderr('Error: Sopel requires Python 3.8+.')
    sys.exit(1)

LOGGER = logging.getLogger(__name__)

ERR_CODE = 1
"""Error code: program exited with an error"""
ERR_CODE_NO_RESTART = 2
"""Error code: program exited with an error and should not be restarted

This error code is used to prevent systemd from restarting the bot when it
encounters such an error case.
"""


def run(settings, pid_file, daemon=False):
    """Run the bot with these ``settings``.

    :param settings: settings with which to run the bot
    :type settings: :class:`sopel.config.Config`
    :param str pid_file: path to the bot's PID file
    :param bool daemon: tell if the bot should be run as a daemon
    """
    delay = 20

    # Acts as a welcome message, showing the program and platform version at start
    print_version()
    # Also show the location of the config file used to load settings
    print("\nLoaded config file: {}".format(settings.filename))

    # Define empty variable `p` for bot
    p = None
    while True:
        if p and p.hasquit:  # Check if `hasquit` was set for bot during disconnected phase
            break
        try:
            p = bot.Sopel(settings, daemon=daemon)
            p.setup()
        except KeyboardInterrupt:
            utils.stderr('Bot setup interrupted')
            break
        except Exception:
            # In that case, there is nothing we can do.
            # If the bot can't setup itself, then it won't run.
            # This is a critical case scenario, where the user should have
            # direct access to the exception traceback right in the console.
            # Besides, we can't know if logging has been set up or not, so
            # we can't rely on that here.
            utils.stderr('Unexpected error in bot setup')
            raise

        try:
            p.run(settings.core.host, int(settings.core.port))
        except KeyboardInterrupt:
            break
        except Exception:
            err_log = logging.getLogger('sopel.exceptions')
            err_log.exception('Critical exception in core')
            err_log.error('----------------------------------------')
            return ERR_CODE

        if p.wantsrestart:
            return -1
        if p.hasquit:
            return 0

        LOGGER.warning('Disconnected. Reconnecting in %s seconds...', delay)
        time.sleep(delay)


def build_parser():
    """Build an argument parser for the bot.

    :return: the argument parser
    :rtype: :class:`argparse.ArgumentParser`
    """
    parser = argparse.ArgumentParser(description='Sopel IRC Bot')

    parser.add_argument('-V', '--version', action='store_true',
                        dest='version',
                        help='Show version number and exit')

    subparsers = parser.add_subparsers(
        title='subcommands',
        description='List of Sopel\'s subcommands',
        dest='action',
        metavar='{start,configure,stop,restart}')

    # manage `start` subcommand
    parser_start = subparsers.add_parser(
        'start',
        description='Start a Sopel instance. '
                    'This command requires an existing configuration file '
                    'that can be generated with ``sopel configure``.',
        help='Start a Sopel instance')
    parser_start.add_argument(
        '-d', '--fork',
        dest='daemonize',
        action='store_true',
        default=False,
        help='Run Sopel as a daemon (fork). This bot will safely run in the '
             'background. The instance will be named after the name of the '
             'configuration file used to run it. '
             'To stop it, use ``sopel stop`` (with the same configuration).')
    utils.add_common_arguments(parser_start)

    # manage `configure` subcommand
    parser_configure = subparsers.add_parser(
        'configure',
        description='Run the configuration wizard. It can be used to create '
                    'a new configuration file or to update an existing one.',
        help='Sopel\'s Wizard tool')
    parser_configure.add_argument(
        '--plugins',
        action='store_true',
        default=False,
        dest='plugins',
        help='Check for Sopel plugins that require configuration, and run '
             'their configuration wizards.')
    utils.add_common_arguments(parser_configure)

    # manage `stop` subcommand
    parser_stop = subparsers.add_parser(
        'stop',
        description='Stop a running Sopel instance. '
                    'This command determines the instance to quit by the name '
                    'of the configuration file used ("default", or the one '
                    'from the ``-c``/``--config`` option). '
                    'This command should be used when the bot is running in '
                    'the background from ``sopel start -d``, and should not '
                    'be used when Sopel is managed by a process manager '
                    '(like systemd or supervisor).',
        help='Stop a running Sopel instance')
    parser_stop.add_argument(
        '-k', '--kill',
        action='store_true',
        default=False,
        help='Kill Sopel without a graceful quit')
    utils.add_common_arguments(parser_stop)

    # manage `restart` subcommand
    parser_restart = subparsers.add_parser(
        'restart',
        description='Restart a running Sopel instance',
        help='Restart a running Sopel instance')
    utils.add_common_arguments(parser_restart)

    return parser


def check_not_root():
    """Check if root is running the bot.

    It raises a ``RuntimeError`` if the user has root privileges on Linux or
    if it is the ``Administrator`` account on Windows.
    """
    opersystem = platform.system()
    if opersystem in ["Linux", "Darwin"]:
        # Linux/Mac
        if os.getuid() == 0 or os.geteuid() == 0:
            raise RuntimeError('Error: Do not run Sopel with root privileges.')
    elif opersystem in ["Windows"]:
        # Windows
        if os.environ.get("USERNAME") == "Administrator":
            raise RuntimeError('Error: Do not run Sopel as Administrator.')
    else:
        utils.stderr(
            "Warning: %s is an uncommon operating system platform. "
            "Sopel should still work, but please contact Sopel's developers "
            "if you experience issues."
            % opersystem)


def print_version():
    """Print Python version and Sopel version on stdout."""
    py_ver = '%s.%s.%s' % (sys.version_info.major,
                           sys.version_info.minor,
                           sys.version_info.micro)
    print('Sopel %s (running on Python %s)' % (__version__, py_ver))
    print('https://sopel.chat/')


def get_configuration(options):
    """Get or create a configuration object from ``options``.

    :param options: argument parser's options
    :type options: :class:`argparse.Namespace`
    :return: a configuration object
    :rtype: :class:`sopel.config.Config`

    This may raise a :exc:`sopel.config.ConfigurationError` if the
    configuration file is invalid.

    .. seealso::

       The configuration file is loaded by
       :func:`~sopel.cli.run.utils.load_settings`

    """
    try:
        settings = utils.load_settings(options)
    except config.ConfigurationNotFound as error:
        raise config.ConfigurationError(
            "%s\n"
            "If you're just setting up Sopel, welcome! "
            "You can use `sopel configure` to get started easily." % error
        )

    settings._is_daemonized = options.daemonize
    return settings


def get_pid_filename(settings, pid_dir):
    """Get the pid file name in ``pid_dir`` from the given ``settings``.

    :param settings: Sopel config
    :type settings: :class:`sopel.config.Config`
    :param str pid_dir: path to the pid directory
    :return: absolute filename of the pid file

    By default, it's ``sopel.pid``, but if the configuration's basename is not
    ``default`` then it will be used to generate the pid file name as
    ``sopel-{basename}.pid`` instead.
    """
    name = 'sopel.pid'
    if settings.basename != 'default':
        filename = os.path.basename(settings.filename)
        basename, ext = os.path.splitext(filename)
        if ext != '.cfg':
            basename = filename
        name = 'sopel-%s.pid' % basename

    return os.path.abspath(os.path.join(pid_dir, name))


def get_running_pid(filename):
    """Retrieve the PID number from the given ``filename``.

    :param str filename: path to file to read the PID from
    :return: the PID number of a Sopel instance if running, ``None`` otherwise
    :rtype: integer

    This function tries to retrieve a PID number from the given ``filename``,
    as an integer, and returns ``None`` if the file is not found or if the
    content is not an integer.
    """
    if not os.path.isfile(filename):
        return

    with open(filename, 'r') as pid_file:
        try:
            return int(pid_file.read())
        except ValueError:
            pass


def command_start(opts):
    """Start a Sopel instance.

    :param opts: parsed arguments
    :type opts: :class:`argparse.Namespace`
    """
    # Step One: Get the configuration file and prepare to run
    try:
        settings = get_configuration(opts)
    except config.ConfigurationError as e:
        utils.stderr(e)
        return ERR_CODE_NO_RESTART

    if settings.core.not_configured:
        utils.stderr('Bot is not configured, can\'t start')
        return ERR_CODE_NO_RESTART

    # Step Two: Handle process-lifecycle options and manage the PID file
    pid_dir = settings.core.pid_dir
    pid_file_path = get_pid_filename(settings, pid_dir)
    pid = get_running_pid(pid_file_path)

    if pid is not None and utils.check_pid(pid):
        utils.stderr('There\'s already a Sopel instance running '
                     'with this config file.')
        utils.stderr('Try using either the `sopel stop` '
                     'or the `sopel restart` command.')
        return ERR_CODE

    if opts.daemonize:
        child_pid = os.fork()
        if child_pid != 0:
            return

    with open(pid_file_path, 'w') as pid_file:
        pid_file.write(str(os.getpid()))

    try:
        # Step Three: Run Sopel
        ret = run(settings, pid_file_path)
    finally:
        # Step Four: Shutdown Clean-Up
        os.unlink(pid_file_path)

    if ret == -1:
        # Restart
        os.execv(sys.executable, ['python'] + sys.argv)
    else:
        # Quit
        return ret


def command_configure(opts):
    """Sopel Configuration Wizard.

    :param opts: parsed arguments
    :type opts: :class:`argparse.Namespace`
    """
    configpath = utils.find_config(opts.configdir, opts.config)
    if opts.plugins:
        utils.plugins_wizard(configpath)
    else:
        utils.wizard(configpath)


def command_stop(opts):
    """Stop a running Sopel instance.

    :param opts: parsed arguments
    :type opts: :class:`argparse.Namespace`
    """
    # Get Configuration
    try:
        settings = utils.load_settings(opts)
    except config.ConfigurationNotFound as error:
        utils.stderr('Configuration "%s" not found' % error.filename)
        return ERR_CODE

    if settings.core.not_configured:
        utils.stderr('Sopel is not configured, can\'t stop')
        return ERR_CODE

    # Configure logging
    logger.setup_logging(settings)

    # Get Sopel's PID
    filename = get_pid_filename(settings, settings.core.pid_dir)
    pid = get_running_pid(filename)

    if pid is None or not utils.check_pid(pid):
        utils.stderr('Sopel is not running!')
        return ERR_CODE

    # Stop Sopel
    if opts.kill:
        utils.stderr('Killing the Sopel')
        os.kill(pid, signal.SIGKILL)
        return

    utils.stderr('Signaling Sopel to stop gracefully')
    if hasattr(signal, 'SIGUSR1'):
        os.kill(pid, signal.SIGUSR1)
    else:
        # Windows will not generate SIGTERM itself
        # https://docs.microsoft.com/en-us/cpp/c-runtime-library/reference/signal
        os.kill(pid, signal.SIGTERM)


def command_restart(opts):
    """Restart a running Sopel instance.

    :param opts: parsed arguments
    :type opts: :class:`argparse.Namespace`
    """
    # Get Configuration
    try:
        settings = utils.load_settings(opts)
    except config.ConfigurationNotFound as error:
        utils.stderr('Configuration "%s" not found' % error.filename)
        return ERR_CODE

    if settings.core.not_configured:
        utils.stderr('Sopel is not configured, can\'t stop')
        return ERR_CODE

    # Configure logging
    logger.setup_logging(settings)

    # Get Sopel's PID
    filename = get_pid_filename(settings, settings.core.pid_dir)
    pid = get_running_pid(filename)

    if pid is None or not utils.check_pid(pid):
        utils.stderr('Sopel is not running!')
        return ERR_CODE

    utils.stderr('Asking Sopel to restart')
    if hasattr(signal, 'SIGUSR2'):
        os.kill(pid, signal.SIGUSR2)
    else:
        # Windows will not generate SIGILL itself
        # https://docs.microsoft.com/en-us/cpp/c-runtime-library/reference/signal
        os.kill(pid, signal.SIGILL)


def main(argv=None):
    """Sopel run script entry point.

    :param list argv: command line arguments
    """
    # Build parser and handle default command
    global_options = ['-h', '--help', '-V', '--version']
    parser = build_parser()

    argv = argv or sys.argv[1:]
    if not argv:
        # No argument: assume start sub-command
        argv = ['start']

    elif argv[0].startswith('-') and argv[0] not in global_options:
        # No sub-command and no global option
        argv = ['start'] + argv

    # Parse The Command Line
    opts = parser.parse_args(argv)

    # Handle "-V/--version" option
    if opts.version:
        print_version()
        return

    try:
        # Check "Do not run as root"
        check_not_root()

        # Select command
        action = getattr(opts, 'action')
        command = {
            'start': command_start,
            'configure': command_configure,
            'stop': command_stop,
            'restart': command_restart,
        }[action]

        # Run command
        return command(opts)
    except KeyError:
        parser.print_usage()
        return ERR_CODE
    except KeyboardInterrupt:
        print("\n\nInterrupted")
        return ERR_CODE
    except RuntimeError as err:
        utils.stderr(str(err))
        return ERR_CODE


if __name__ == '__main__':
    sys.exit(main())



"""
adminchannel.py - Sopel Channel Admin Plugin
Copyright 2010-2011, Michael Yanovich, Alek Rollyson, and Elsie Powell
Copyright © 2012, Elad Alfassa <elad@fedoraproject.org>
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

import re

from sopel import formatting, plugin


ERROR_MESSAGE_NOT_OP = "I'm not a channel operator!"
ERROR_MESSAGE_NO_PRIV = "You are not a channel operator."


def default_mask(trigger):
    welcome = formatting.color('Welcome to:', formatting.colors.PURPLE)
    chan = formatting.color(trigger.sender, formatting.colors.TEAL)
    topic_ = formatting.bold('Topic:')
    topic_ = formatting.color('| ' + topic_, formatting.colors.PURPLE)
    arg = formatting.color('{}', formatting.colors.GREEN)
    return '{} {} {} {}'.format(welcome, chan, topic_, arg)


@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP, ERROR_MESSAGE_NO_PRIV, reply=True)
@plugin.require_bot_privilege(plugin.OP, ERROR_MESSAGE_NOT_OP, reply=True)
@plugin.command('op')
def op(bot, trigger):
    """
    Command to op users in a room. If no nick is given,
    Sopel will op the nick who sent the command
    """
    nick = trigger.group(2)
    channel = trigger.sender
    if not nick:
        nick = trigger.nick
    bot.write(['MODE', channel, "+o", nick])


@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP, ERROR_MESSAGE_NO_PRIV, reply=True)
@plugin.require_bot_privilege(plugin.OP, ERROR_MESSAGE_NOT_OP, reply=True)
@plugin.command('deop')
def deop(bot, trigger):
    """
    Command to deop users in a room. If no nick is given,
    Sopel will deop the nick who sent the command
    """
    nick = trigger.group(2)
    channel = trigger.sender
    if not nick:
        nick = trigger.nick
    bot.write(['MODE', channel, "-o", nick])


@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP, ERROR_MESSAGE_NO_PRIV, reply=True)
@plugin.require_bot_privilege(plugin.HALFOP, ERROR_MESSAGE_NOT_OP, reply=True)
@plugin.command('voice')
def voice(bot, trigger):
    """
    Command to voice users in a room. If no nick is given,
    Sopel will voice the nick who sent the command
    """
    nick = trigger.group(2)
    channel = trigger.sender
    if not nick:
        nick = trigger.nick
    bot.write(['MODE', channel, "+v", nick])


@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP, ERROR_MESSAGE_NO_PRIV, reply=True)
@plugin.require_bot_privilege(plugin.HALFOP, ERROR_MESSAGE_NOT_OP, reply=True)
@plugin.command('devoice')
def devoice(bot, trigger):
    """
    Command to devoice users in a room. If no nick is given,
    Sopel will devoice the nick who sent the command
    """
    nick = trigger.group(2)
    channel = trigger.sender
    if not nick:
        nick = trigger.nick
    bot.write(['MODE', channel, "-v", nick])


@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP, ERROR_MESSAGE_NO_PRIV, reply=True)
@plugin.require_bot_privilege(plugin.HALFOP, ERROR_MESSAGE_NOT_OP, reply=True)
@plugin.command('kick')
@plugin.priority('high')
def kick(bot, trigger):
    """Kick a user from the channel."""
    text = trigger.group().split()
    argc = len(text)
    if argc < 2:
        return
    opt = bot.make_identifier(text[1])
    nick = opt
    channel = trigger.sender
    reasonidx = 2
    if not opt.is_nick():
        if argc < 3:
            return
        nick = text[2]
        channel = opt
        reasonidx = 3
    reason = ' '.join(text[reasonidx:])

    if nick == bot.nick:
        bot.reply("Hey! Don't kick me. :(")
        return

    bot.kick(nick, channel, reason)


def configureHostMask(mask):
    # shortcut for nick!*@*
    if re.match(r'^[^.@!/\s]+$', mask) is not None:
        return '%s!*@*' % mask

    # shortcut for *!*@host
    # won't work for local names w/o dot, but does support cloaks/with/slashes
    if re.match(r'^[^@!\s]+$', mask) is not None:
        return '*!*@%s' % mask

    # shortcut for *!user@*
    # requires trailing @ to be recognized as a username instead of a nick
    m = re.match(r'^([^!@\s]+)@$', mask)
    if m is not None:
        return '*!%s@*' % m.group(1)

    # shortcut for *!user@host
    m = re.match(r'^([^!@\s]+)@([^@!\s]+)$', mask)
    if m is not None:
        return '*!%s@%s' % (m.group(1), m.group(2))

    # shortcut for nick!user@*
    m = re.match(r'^([^!@\s]+)!([^!@\s]+)@?$', mask)
    if m is not None:
        return '%s!%s@*' % (m.group(1), m.group(2))

    # not a shortcut; validate full NUH format
    if re.match(r'^[^!@\s]+![^!@\s]+@[^!@\s]+$', mask) is not None:
        return mask

    # not a shortcut nor a valid hostmask
    raise ValueError('Invalid hostmask format or unsupported shorthand')


@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP, ERROR_MESSAGE_NO_PRIV, reply=True)
@plugin.require_bot_privilege(plugin.HALFOP, ERROR_MESSAGE_NOT_OP, reply=True)
@plugin.command('ban')
@plugin.priority('high')
def ban(bot, trigger):
    """Ban a user from the channel

    The bot must be a channel operator for this command to work.
    """
    text = trigger.group().split()
    argc = len(text)
    if argc < 2:
        return
    opt = bot.make_identifier(text[1])
    banmask = opt
    channel = trigger.sender
    if not opt.is_nick():
        if argc < 3:
            return
        channel = opt
        banmask = text[2]

    try:
        banmask = configureHostMask(banmask)
    except ValueError:
        return

    bot.write(['MODE', channel, '+b', banmask])


@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP, ERROR_MESSAGE_NO_PRIV, reply=True)
@plugin.require_bot_privilege(plugin.HALFOP, ERROR_MESSAGE_NOT_OP, reply=True)
@plugin.command('unban')
def unban(bot, trigger):
    """Unban a user from the channel

    The bot must be a channel operator for this command to work.
    """
    text = trigger.group().split()
    argc = len(text)
    if argc < 2:
        return
    opt = bot.make_identifier(text[1])
    banmask = opt
    channel = trigger.sender
    if not opt.is_nick():
        if argc < 3:
            return
        channel = opt
        banmask = text[2]

    try:
        banmask = configureHostMask(banmask)
    except ValueError:
        return

    bot.write(['MODE', channel, '-b', banmask])


@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP, ERROR_MESSAGE_NO_PRIV, reply=True)
@plugin.require_bot_privilege(plugin.OP, ERROR_MESSAGE_NOT_OP, reply=True)
@plugin.command('quiet')
def quiet(bot, trigger):
    """Quiet a user

    The bot must be a channel operator for this command to work.
    """
    text = trigger.group().split()
    argc = len(text)
    if argc < 2:
        return
    opt = bot.make_identifier(text[1])
    quietmask = opt
    channel = trigger.sender
    if not opt.is_nick():
        if argc < 3:
            return
        quietmask = text[2]
        channel = opt

    try:
        quietmask = configureHostMask(quietmask)
    except ValueError:
        return

    bot.write(['MODE', channel, '+q', quietmask])


@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP, ERROR_MESSAGE_NO_PRIV, reply=True)
@plugin.require_bot_privilege(plugin.OP, ERROR_MESSAGE_NOT_OP, reply=True)
@plugin.command('unquiet')
def unquiet(bot, trigger):
    """Unquiet a user

    The bot must be a channel operator for this command to work.
    """
    text = trigger.group().split()
    argc = len(text)
    if argc < 2:
        return
    opt = bot.make_identifier(text[1])
    quietmask = opt
    channel = trigger.sender
    if not opt.is_nick():
        if argc < 3:
            return
        quietmask = text[2]
        channel = opt

    try:
        quietmask = configureHostMask(quietmask)
    except ValueError:
        return

    bot.write(['MODE', channel, '-q', quietmask])


@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP, ERROR_MESSAGE_NO_PRIV, reply=True)
@plugin.require_bot_privilege(plugin.OP, ERROR_MESSAGE_NOT_OP, reply=True)
@plugin.command('kickban', 'kb')
@plugin.example('.kickban [#chan] user1 user!*@* get out of here')
@plugin.priority('high')
def kickban(bot, trigger):
    """Kick and ban a user from the channel

    The bot must be a channel operator for this command to work.
    """
    text = trigger.group().split()
    argc = len(text)
    if argc < 4:
        return
    opt = bot.make_identifier(text[1])
    nick = opt
    mask = text[2]
    channel = trigger.sender
    reasonidx = 3
    if not opt.is_nick():
        if argc < 5:
            return
        channel = opt
        nick = text[2]
        mask = text[3]
        reasonidx = 4
    reason = ' '.join(text[reasonidx:])

    try:
        mask = configureHostMask(mask)
    except ValueError:
        return

    bot.write(['MODE', channel, '+b', mask])
    bot.kick(nick, channel, reason)


@plugin.require_chanmsg
@plugin.command('topic')
def topic(bot, trigger):
    """Change the channel topic

    The bot must be a channel operator for this command to work in +t channels.
    """
    mode_t = bot.channels[trigger.sender].modes.get("t", False)
    if mode_t and not bot.has_channel_privilege(trigger.sender, plugin.HALFOP):
        bot.reply(ERROR_MESSAGE_NOT_OP)
        return
    if mode_t and bot.channels[trigger.sender].privileges[trigger.nick] < plugin.HALFOP:
        bot.reply(ERROR_MESSAGE_NO_PRIV)
        return
    if not trigger.group(2):
        return
    channel = trigger.sender.lower()

    mask = bot.db.get_channel_value(
        channel, 'topic_mask', default_mask(trigger))
    mask = mask.replace('%s', '{}')
    narg = len(re.findall('{}', mask))

    top = trigger.group(2)
    args = []
    if top:
        args = top.split('~', narg)

    if len(args) != narg:
        message = "Not enough arguments. You gave {}; it requires {}.".format(
            len(args), narg)
        bot.reply(message)
        return
    topic = mask.format(*args)

    topiclen = getattr(bot.isupport, 'TOPICLEN', None)
    if topiclen is not None:
        my_len = len(topic.encode('utf-8'))
        if my_len > topiclen:
            bot.reply(
                "Formatted topic is too long ({} bytes); "
                "the server limit is {} bytes."
                .format(my_len, topiclen))
            return

    bot.write(('TOPIC', channel + ' :' + topic))


@plugin.require_privilege(plugin.OP, ERROR_MESSAGE_NO_PRIV)
@plugin.commands('tmask set', 'tmask get', 'tmask clear', 'tmask')
@plugin.example('.tmask clear', user_help=True)
@plugin.example('.tmask get', user_help=True)
@plugin.example('.tmask set My {} topic mask!', user_help=True)
def topic_mask_management(bot, trigger):
    """Set, get, or clear the current channel's topic mask.

    Recognized subcommands are 'set', 'get', and 'clear'. A plain 'tmask'
    command with no arguments is equivalent to 'tmask get'.

    This mask is used by the 'topic' command. `{}` allows interpolating a chunk
    of text within the topic mask template.
    """
    command, _, subcommand = trigger.group(1).partition(' ')

    if not subcommand or subcommand == 'get':
        mask = bot.db.get_channel_value(
            trigger.sender, 'topic_mask', default_mask(trigger))
        bot.reply('Current topic mask: {}'.format(mask))
        return

    if subcommand == 'set':
        if not trigger.group(2):
            message = (
                'I need a non-empty topic mask to set. '
                'To delete the saved topic mask, use `{prefix}tmask clear`.'
            ).format(prefix=bot.settings.core.help_prefix)
            bot.reply(message)
            return

        bot.db.set_channel_value(trigger.sender, 'topic_mask', trigger.group(2))
        message = (
            'Topic mask set. '
            'Use `{prefix}topic <args>` to set topic, '
            '`{prefix}tmask get` to see the current mask, '
            'and `{prefix}tmask clear` to delete the saved topic mask.'
        ).format(prefix=bot.settings.core.help_prefix)
        bot.reply(message)
        return

    if subcommand == 'clear':
        bot.db.delete_channel_value(trigger.sender, 'topic_mask')
        bot.reply('Cleared topic mask.')
        return



"""
uptime.py - Sopel Uptime Plugin
Copyright 2014, Fabian Neundorf
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

from datetime import datetime, timedelta, timezone

from sopel import plugin


def setup(bot):
    if "start_time" not in bot.memory:
        bot.memory["start_time"] = datetime.now(timezone.utc)


@plugin.command('uptime')
@plugin.example('.uptime', user_help=True)
@plugin.output_prefix('[uptime] ')
def uptime(bot, trigger):
    """Return the uptime of Sopel."""
    delta = timedelta(seconds=round((datetime.now(timezone.utc) -
                                    bot.memory["start_time"])
                                    .total_seconds()))
    bot.say("I've been sitting here for {} and I keep going!".format(delta))



"""User and channel objects used in state tracking."""
from __future__ import annotations

import functools
from typing import Any, Optional, TYPE_CHECKING, Union

from sopel.privileges import AccessLevel
from sopel.tools import memories
from sopel.tools.identifiers import Identifier, IdentifierFactory

if TYPE_CHECKING:
    import datetime


@functools.total_ordering
class User:
    """A representation of a user Sopel is aware of.

    :param nick: the user's nickname
    :type nick: :class:`sopel.tools.identifiers.Identifier`
    :param str user: the user's local username ("user" in `user@host.name`)
    :param str host: the user's hostname ("host.name" in `user@host.name`)
    """
    __slots__ = (
        'nick', 'user', 'host', 'realname', 'channels', 'account', 'away', 'is_bot',
    )

    def __init__(
        self,
        nick: Identifier,
        user: Optional[str],
        host: Optional[str],
    ) -> None:
        assert isinstance(nick, Identifier)
        self.nick: Identifier = nick
        """The user's nickname."""
        self.user: Optional[str] = user
        """The user's local username.

        Will be ``None`` if Sopel has not yet received complete user
        information from the IRC server.
        """
        self.host: Optional[str] = host
        """The user's hostname.

        Will be ``None`` if Sopel has not yet received complete user
        information from the IRC server.
        """
        self.realname: Optional[str] = None
        """The user's realname.

        Will be ``None`` if Sopel has not yet received complete user
        information from the IRC server.
        """
        self.channels: dict[Identifier, 'Channel'] = {}
        """The channels the user is in.

        This maps channel name :class:`~sopel.tools.identifiers.Identifier`\\s
        to :class:`Channel` objects.
        """
        self.account: Optional[str] = None
        """The IRC services account of the user.

        This relies on IRCv3 account tracking being enabled.

        Will be ``None`` if the user is not logged into an account (including
        when account tracking is not supported by the IRC server.)
        """
        self.away: Optional[bool] = None
        """Whether the user is marked as away.

        Will be ``None`` if the user's current away state hasn't been
        established yet (via WHO or other means such as ``away-notify``).
        """
        self.is_bot: Optional[bool] = None
        """Whether the user is flagged as a bot.

        Will be ``None`` if the user hasn't yet been WHOed, or if the IRC
        server does not support a 'bot' user mode.
        """

    @property
    def hostmask(self) -> str:
        """The user's full hostmask (``nick!user@host``)."""
        # TODO: this won't work as expected if `user`/`host` is still `None`
        return '{}!{}@{}'.format(
            self.nick,
            self.user,
            self.host,
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, User):
            return NotImplemented
        return self.nick == other.nick

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, User):
            return NotImplemented
        return self.nick < other.nick


@functools.total_ordering
class Channel:
    """A representation of a channel Sopel is in.

    :param name: the channel name
    :type name: :class:`~sopel.tools.identifiers.Identifier`
    :param identifier_factory: A factory to create
                               :class:`~sopel.tools.identifiers.Identifier`\\s
    """
    __slots__ = (
        'name',
        'users',
        'privileges',
        'topic',
        'modes',
        'last_who',
        'join_time',
        'make_identifier',
    )

    def __init__(
        self,
        name: Identifier,
        identifier_factory: IdentifierFactory = Identifier,
    ) -> None:
        assert isinstance(name, Identifier)
        self.name: Identifier = name
        """The name of the channel."""

        self.make_identifier: IdentifierFactory = identifier_factory
        """Factory to create :class:`~sopel.tools.identifiers.Identifier`.

        ``Identifier`` is used for :class:`User`'s nick, and the channel
        needs to translate nicks from string to ``Identifier`` when
        manipulating data associated to a user by its nickname.
        """

        self.users: dict[
            Identifier,
            User,
        ] = memories.SopelIdentifierMemory(
            identifier_factory=self.make_identifier,
        )
        """The users in the channel.

        This maps nickname :class:`~sopel.tools.identifiers.Identifier`\\s to
        :class:`User` objects.
        """
        self.privileges: dict[
            Identifier,
            int,
        ] = memories.SopelIdentifierMemory(
            identifier_factory=self.make_identifier,
        )
        """The permissions of the users in the channel.

        This maps nickname :class:`~sopel.tools.identifiers.Identifier`\\s to
        bitwise integer values. This can be compared to appropriate constants
        from :class:`sopel.privileges.AccessLevel`.
        """
        self.topic: str = ''
        """The topic of the channel."""

        self.modes: dict[str, Union[set, str, bool]] = {}
        """The channel's modes.

        For type A modes (nick/address list), the value is a set. For type B
        (parameter) or C (parameter when setting), the value is a string. For
        type D, the value is ``True``.

        .. note::

            Type A modes may only contain changes the bot has observed. Sopel
            does not automatically populate all modes and lists.
        """

        self.last_who: Optional[datetime.datetime] = None
        """The last time a WHO was requested for the channel."""

        self.join_time: Optional[datetime.datetime] = None
        """The time the server acknowledged our JOIN message.

        Based on server-reported time if the ``server-time`` IRCv3 capability
        is available, otherwise the time Sopel received it.
        """

    def clear_user(self, nick: Identifier) -> None:
        """Remove ``nick`` from this channel.

        :param nick: the nickname of the user to remove

        Called after a user leaves the channel via PART, KICK, QUIT, etc.
        """
        user = self.users.pop(nick, None)
        self.privileges.pop(nick, None)
        if user is not None:
            user.channels.pop(self.name, None)

    def add_user(self, user: User, privs: int = 0) -> None:
        """Add ``user`` to this channel.

        :param user: the new user to add
        :param privs: privilege bitmask (see constants in
                      :class:`sopel.privileges.AccessLevel`)

        Called when a new user JOINs the channel.
        """
        assert isinstance(user, User)
        self.users[user.nick] = user
        self.privileges[user.nick] = privs or 0
        user.channels[self.name] = self

    def has_privilege(self, nick: str, privilege: int) -> bool:
        """Tell if a user has a ``privilege`` level or above in this channel.

        :param nick: a user's nick in this channel
        :param privilege: privilege level to check

        This method checks the user's privilege level in this channel, i.e. if
        it has this level or higher privileges::

            >>> channel.add_user(some_user, plugin.OP)
            >>> channel.has_privilege(some_user.nick, plugin.VOICE)
            True

        The ``nick`` argument can be either a :class:`str` or a
        :class:`sopel.tools.identifiers.Identifier`. If the user is not in this
        channel, it will be considered as not having any privilege.

        .. seealso::

            There are other methods to check the exact privilege level of a
            user, such as :meth:`is_oper`, :meth:`is_owner`, :meth:`is_admin`,
            :meth:`is_op`, :meth:`is_halfop`, and :meth:`is_voiced`.

        .. important::

            Not all IRC networks support all privilege levels. If you intend
            for your plugin to run on any network, it is safest to rely only
            on the presence of standard modes: ``+v`` (voice) and ``+o`` (op).

        """
        return self.privileges.get(self.make_identifier(nick), 0) >= privilege

    def is_oper(self, nick: str) -> bool:
        """Tell if a user has the OPER (operator) privilege level.

        :param nick: a user's nick in this channel

        Unlike :meth:`has_privilege`, this method checks if the user has been
        explicitly granted the OPER privilege level::

            >>> channel.add_user(some_user, plugin.OPER)
            >>> channel.is_oper(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            False

        Note that you can always have more than one privilege level::

            >>> channel.add_user(some_user, plugin.OPER | plugin.VOICE)
            >>> channel.is_oper(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            True

        .. important::

            Not all IRC networks support this privilege mode. If you are
            writing a plugin for public distribution, ensure your code behaves
            sensibly if only ``+v`` (voice) and ``+o`` (op) modes exist.

        """
        identifier = self.make_identifier(nick)
        return bool(self.privileges.get(identifier, 0) & AccessLevel.OPER)

    def is_owner(self, nick: str) -> bool:
        """Tell if a user has the OWNER privilege level.

        :param nick: a user's nick in this channel

        Unlike :meth:`has_privilege`, this method checks if the user has been
        explicitly granted the OWNER privilege level::

            >>> channel.add_user(some_user, plugin.OWNER)
            >>> channel.is_owner(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            False

        Note that you can always have more than one privilege level::

            >>> channel.add_user(some_user, plugin.OWNER | plugin.VOICE)
            >>> channel.is_owner(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            True

        .. important::

            Not all IRC networks support this privilege mode. If you are
            writing a plugin for public distribution, ensure your code behaves
            sensibly if only ``+v`` (voice) and ``+o`` (op) modes exist.

        """
        identifier = self.make_identifier(nick)
        return bool(self.privileges.get(identifier, 0) & AccessLevel.OWNER)

    def is_admin(self, nick: str) -> bool:
        """Tell if a user has the ADMIN privilege level.

        :param nick: a user's nick in this channel

        Unlike :meth:`has_privilege`, this method checks if the user has been
        explicitly granted the ADMIN privilege level::

            >>> channel.add_user(some_user, plugin.ADMIN)
            >>> channel.is_admin(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            False

        Note that you can always have more than one privilege level::

            >>> channel.add_user(some_user, plugin.ADMIN | plugin.VOICE)
            >>> channel.is_admin(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            True

        .. important::

            Not all IRC networks support this privilege mode. If you are
            writing a plugin for public distribution, ensure your code behaves
            sensibly if only ``+v`` (voice) and ``+o`` (op) modes exist.

        """
        identifier = self.make_identifier(nick)
        return bool(self.privileges.get(identifier, 0) & AccessLevel.ADMIN)

    def is_op(self, nick: str) -> bool:
        """Tell if a user has the OP privilege level.

        :param nick: a user's nick in this channel

        Unlike :meth:`has_privilege`, this method checks if the user has been
        explicitly granted the OP privilege level::

            >>> channel.add_user(some_user, plugin.OP)
            >>> channel.is_op(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            False

        Note that you can always have more than one privilege level::

            >>> channel.add_user(some_user, plugin.OP | plugin.VOICE)
            >>> channel.is_op(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            True

        """
        identifier = self.make_identifier(nick)
        return bool(self.privileges.get(identifier, 0) & AccessLevel.OP)

    def is_halfop(self, nick: str) -> bool:
        """Tell if a user has the HALFOP privilege level.

        :param nick: a user's nick in this channel

        Unlike :meth:`has_privilege`, this method checks if the user has been
        explicitly granted the HALFOP privilege level::

            >>> channel.add_user(some_user, plugin.HALFOP)
            >>> channel.is_halfop(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            False

        Note that you can always have more than one privilege level::

            >>> channel.add_user(some_user, plugin.HALFOP | plugin.VOICE)
            >>> channel.is_halfop(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            True

        .. important::

            Not all IRC networks support this privilege mode. If you are
            writing a plugin for public distribution, ensure your code behaves
            sensibly if only ``+v`` (voice) and ``+o`` (op) modes exist.

        """
        identifier = self.make_identifier(nick)
        return bool(self.privileges.get(identifier, 0) & AccessLevel.HALFOP)

    def is_voiced(self, nick: str) -> bool:
        """Tell if a user has the VOICE privilege level.

        :param nick: a user's nick in this channel

        Unlike :meth:`has_privilege`, this method checks if the user has been
        explicitly granted the VOICE privilege level::

            >>> channel.add_user(some_user, plugin.VOICE)
            >>> channel.is_voiced(some_user.nick)
            True
            >>> channel.add_user(some_user, plugin.OP)
            >>> channel.is_voiced(some_user.nick)
            False

        Note that you can always have more than one privilege level::

            >>> channel.add_user(some_user, plugin.VOICE | plugin.OP)
            >>> channel.is_voiced(some_user.nick)
            True
            >>> channel.is_op(some_user.nick)
            True

        """
        identifier = self.make_identifier(nick)
        return bool(self.privileges.get(identifier, 0) & AccessLevel.VOICE)

    def rename_user(self, old: Identifier, new: Identifier) -> None:
        """Rename a user.

        :param old: the user's old nickname
        :param new: the user's new nickname

        Called on ``NICK`` events.
        """
        if old in self.users:
            self.users[new] = self.users.pop(old)
            self.users[new].nick = new
        if old in self.privileges:
            self.privileges[new] = self.privileges.pop(old)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Channel):
            return NotImplemented
        return self.name == other.name

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, Channel):
            return NotImplemented
        return self.name < other.name



"""Test tools, factories, pytest fixtures, and mocks.

.. versionadded:: 7.0
"""
from __future__ import annotations


def rawlist(*args):
    """Build a list of raw IRC messages from the lines given as ``*args``.

    :return: a list of raw IRC messages as seen by the bot
    :rtype: list

    This is a helper function to build a list of messages without having to
    care about encoding or this pesky carriage return::

        >>> rawlist('PRIVMSG :Hello!')
        [b'PRIVMSG :Hello!\\r\\n']
    """
    return ['{0}\r\n'.format(arg).encode('utf-8') for arg in args]



"""Sopel Config Command Line Interface (CLI): ``sopel-config``"""
from __future__ import annotations

import argparse
import os

from . import utils


def build_parser():
    """Build and configure an argument parser for ``sopel-config``.

    :return: the argument parser
    :rtype: :class:`argparse.ArgumentParser`
    """
    parser = argparse.ArgumentParser(
        description='Sopel configuration tool')

    # Subparser: sopel-config <sub-parser> <sub-options>
    subparsers = parser.add_subparsers(
        help='Actions to perform',
        dest='action')

    # sopel-config list
    list_parser = subparsers.add_parser(
        'list',
        help="List available configurations from Sopel's config directory",
        description="""
            List available configurations from Sopel's config directory
            with the extension "{ext}". Use option ``--config-dir`` to use a
            specific config directory.
        """.format(ext='.cfg'))
    utils.add_common_arguments(list_parser)
    list_parser.add_argument(
        '-e', '--ext', '--extension',
        dest='extension',
        default='.cfg',
        help='Filter by extension (default to "%(default)s)"')
    list_parser.add_argument(
        '-p', '--path',
        action='store_true',
        dest='display_path',
        default=False,
        help='Display a list of absolute filenames instead of their names')

    # sopel-config init
    init_parser = subparsers.add_parser(
        'init',
        help='Initialize Sopel configuration file',
        description='Initialize Sopel configuration file')
    utils.add_common_arguments(init_parser)

    # sopel-config get <section> <key>
    get_parser = subparsers.add_parser(
        'get',
        help="Get a configuration option's value",
        description="Get a configuration option's value",
    )
    get_parser.add_argument(
        'section',
        help='The name of the section to look in',
    )
    get_parser.add_argument(
        'option',
        help='The name of the option to retrieve',
    )
    utils.add_common_arguments(get_parser)

    return parser


def handle_list(options):
    """Display a list of configurations available in Sopel's config directory.

    :param options: parsed arguments
    :type options: :class:`argparse.Namespace`
    :return: 0 if everything went fine

    This command displays an unordered list of config names from Sopel's
    config directory, without their extensions::

        $ sopel-config list
        default
        custom

    By default, the config directory is ``~/.sopel``. To select a different
    config directory, options ``--config-dir`` can be used.

    It is possible to filter by extension using the
    ``-e``/``--ext``/``--extension`` option; default is ``.cfg``
    (the ``.`` prefix is not required).
    """
    configdir = options.configdir
    display_path = options.display_path
    extension = options.extension
    if not extension.startswith('.'):
        extension = '.' + extension
    configs = utils.enumerate_configs(configdir, extension)

    found = False
    for config_filename in configs:
        found = True
        if display_path:
            print(os.path.join(configdir, config_filename))
        else:
            name, _ = os.path.splitext(config_filename)
            print(name)

    if not found:
        utils.stderr('No config file found at this location: %s' % configdir)
        utils.stderr('Use `sopel-config init` to create a new config file.')

    return 0  # successful operation


def handle_init(options):
    """Use config wizard to initialize a new configuration file for the bot.

    :param options: parsed arguments
    :type options: :class:`argparse.Namespace`
    :return: 0 if everything went fine;
             1 if the file is invalid or if it already exists

    .. note::

       Due to how the config wizard works, the configuration filename's
       extension **must be** ``.cfg``.

    """
    config_filename = utils.find_config(options.configdir, options.config)
    config_name, ext = os.path.splitext(config_filename)

    if ext and ext != '.cfg':
        utils.stderr('Configuration wizard accepts .cfg files only')
        return 1
    elif not ext:
        config_filename = config_name + '.cfg'

    if os.path.isfile(config_filename):
        utils.stderr('Configuration file %s already exists' % config_filename)
        return 1

    print('Starting Sopel config wizard for: %s' % config_filename)
    try:
        utils.wizard(config_name)
    except KeyboardInterrupt:
        utils.stderr('\nOperation cancelled; no file has been created.')
        return 1  # cancelled operation

    return 0  # successful operation


def handle_get(options):
    """Read the settings to display the value of ``<section> <key>``.

    :param options: parsed arguments
    :type options: :class:`argparse.Namespace`
    :return: 0 if everything went fine;
             1 if the section and/or key does not exist;
             2 if the settings can't be loaded
    """
    try:
        settings = utils.load_settings(options)
    except Exception as error:
        utils.stderr(error)
        return 2

    section = options.section
    option = options.option

    # Making sure the section.option exists
    if not settings.parser.has_section(section):
        utils.stderr('Section "%s" does not exist' % section)
        return 1
    if not settings.parser.has_option(section, option):
        utils.stderr(
            'Section "%s" does not have a "%s" option' % (section, option))
        return 1

    # Display the value
    print(settings.get(section, option))
    return 0  # successful operation


def main():
    """Console entry point for ``sopel-config``."""
    parser = build_parser()
    options = parser.parse_args()
    action = options.action

    if not action:
        parser.print_help()
        return

    try:
        # init command does not require existing settings
        if action == 'list':
            return handle_list(options)
        elif action == 'init':
            return handle_init(options)
        elif action == 'get':
            return handle_get(options)
    except KeyboardInterrupt:
        # ctrl+c was used, nothing to report here
        pass



"""This contains decorators and other tools for creating Sopel plugins."""
# Copyright 2013, Ari Koivula, <ari@koivu.la>
# Copyright © 2013, Elad Alfassa <elad@fedoraproject.org>
# Copyright 2013, Lior Ramati <firerogue517@gmail.com>
# Copyright 2019, deathbybandaid, deathbybandaid.net
# Copyright 2019, dgw, technobabbl.es
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
# Licensed under the Eiffel Forum License 2.

from __future__ import annotations

import enum
import functools
import logging
import re
from typing import (
    Callable,
    Literal,
    Optional,
    Pattern,
    Protocol,
    TYPE_CHECKING,
    Union,
)

# import and expose privileges as shortcut
from sopel.privileges import AccessLevel

VOICE = AccessLevel.VOICE
HALFOP = AccessLevel.HALFOP
OP = AccessLevel.OP
ADMIN = AccessLevel.ADMIN
OWNER = AccessLevel.OWNER
OPER = AccessLevel.OPER


if TYPE_CHECKING:
    from collections.abc import Iterable
    from sopel.bot import SopelWrapper

__all__ = [
    # constants
    'NOLIMIT', 'VOICE', 'HALFOP', 'OP', 'ADMIN', 'OWNER', 'OPER',
    # decorators
    'action_command',
    'action_commands',
    'allow_bots',
    'capability',
    'CapabilityNegotiation',
    'command',
    'commands',
    'ctcp',
    'echo',
    'event',
    'example',
    'find',
    'find_lazy',
    'interval',
    'label',
    'nickname_command',
    'nickname_commands',
    'output_prefix',
    'priority',
    'rate',
    'rate_user',
    'rate_channel',
    'rate_global',
    'require_account',
    'require_admin',
    'require_bot_privilege',
    'require_chanmsg',
    'require_owner',
    'require_privilege',
    'require_privmsg',
    'rule',
    'rule_lazy',
    'search',
    'search_lazy',
    'thread',
    'unblockable',
    'url',
    'url_lazy',
]


LOGGER = logging.getLogger(__name__)


NOLIMIT = 1
"""Return value for ``callable``\\s, which suppresses rate limiting.

Returning this value means the triggering user will not be prevented from
triggering the same callable again within the rate limit. This can be used,
for example, to allow a user to retry a failed command immediately.

.. versionadded:: 4.0
"""


class CapabilityNegotiation(enum.Enum):
    """Capability Negotiation status."""

    DONE = enum.auto()
    """The capability negotiation can end.

    This must be returned by a capability request handler to signify to the bot
    that the capability has been properly negotiated and negotiation can end if
    all other conditions are met.
    """

    CONTINUE = enum.auto()
    """The capability negotiation must continue.

    This must be returned by a capability request handler to signify to the bot
    that the capability requires further processing (e.g. SASL
    authentication) and negotiation must not end yet.

    The plugin author MUST signal the bot once the negotiation is done.
    """

    ERROR = enum.auto()
    """The capability negotiation callback was improperly executed.

    If a capability request's handler returns this status, or if it raises an
    exception, the bot will mark the request as errored. A handler can use this
    return value to inform the bot that something wrong happened, without being
    an error in the code itself.
    """


class CapabilityHandler(Protocol):
    """:class:`~typing.Protocol` definition for capability handler.

    When a plugin requests a capability, it can define a callback handler for
    that request using :class:`capability` as a decorator. That handler will be
    called upon Sopel receiving either an ``ACK`` (capability enabled) or a
    ``NAK`` (capability denied) CAP message.

    Example::

        from sopel import plugin
        from sopel.bot import SopelWrapper

        @plugin.capability('example/cap-name')
        def capability_handler(
            cap_req: tuple[str, ...],
            bot: SopelWrapper,
            acknowledged: bool,
        ) -> plugin.CapabilityNegotiation:
            if acknowledged:
                # do something if acknowledged
                # i.e.
                # activate a plugin's feature
                pass
            else:
                # do something else if not
                # i.e. use a fallback mechanism
                # or deactivate a plugin's feature if needed
                pass

            # always return if Sopel can send "CAP END" (DONE)
            # or if the plugin must notify the bot for that later (CONTINUE)
            return plugin.CapabilityNegotiation.DONE

    .. note::

        This protocol class should be used for type checking and documentation
        purposes only.

    """
    def __call__(
        self,
        cap_req: tuple[str, ...],
        bot: SopelWrapper,
        acknowledged: bool,
    ) -> CapabilityNegotiation:
        """A capability handler must be a callable with this signature.

        :param cap_req: the capability request, as a tuple of string
        :param bot: the bot instance
        :param acknowledged: that flag that tells if the capability is enabled
                             or denied
        :return: the return value indicates if the capability negotiation is
                 complete for this request or not
        """


class capability:
    """Decorate a function to request a capability and handle the result.

    :param name: name of the capability to negotiate with the server; this
                 positional argument can be used multiple times to form a
                 single ``CAP REQ``
    :param handler: optional keyword argument, acknowledgement handler

    The Client Capability Negotiation is a feature of IRCv3 that exposes a
    mechanism for a server to advertise a list of features and for clients to
    request them when they are available.

    This decorator will register a capability request, allowing the bot to
    request capabilities if they are available. You can request more than one
    at a time, which will make for one single request.

    The handler must follow the :class:`CapabilityHandler` protocol.

    .. note::

        Due to how Capability Negotiation works, a request will be acknowledged
        or denied all at once. This means that this may succeed::

            @plugin.capability('away-notify')

        But this may not::

            @plugin.capability('away-notify', 'example/incompatible-cap')

        Even though the ``away-notify`` capability is available and can be
        enabled, the second ``CAP REQ`` will be denied because the server won't
        acknowledge a request that contains an incompatible capability.

        In that case, if you don't need both at the same time, you should use
        two different handlers::

            @plugin.capability('away-notify')
            def cap_away_notify(cap_req, bot, ack):
                # handle away-notify acknowledgement

            @plugin.capability('example/incompatible-cap')
            def cap_example_incompatible_cap(cap_req, bot, ack):
                # handle example/incompatible-cap acknowledgement
                # or, most probably, lack thereof

        This will allow the server to acknowledge or deny each capability
        independently.

    .. warning::

        A function cannot be decorated more than once by this decorator, as
        the result is an instance of :class:`capability`.

        If you want to handle a ``CAP`` message without requesting the
        capability, you should use the :func:`event` decorator instead.

    .. warning::

        The list of ``cap_req`` is limited in size to prevent the bot from
        separating the ``CAP REQ`` in multiple lines as the bot does not know
        how to call back the capability handler upon receiving the multi-line
        ``ACK * REQ``.

    .. seealso::

        The IRCv3 specification on `Client Capability Negotiation`__.

    .. __: https://ircv3.net/specs/extensions/capability-negotiation
    """
    def __init__(
        self,
        *cap_req: str,
        handler: Optional[CapabilityHandler] = None,
    ) -> None:
        cap_req_text = ' '.join(cap_req)
        if len(cap_req_text.encode('utf-8')) > 500:
            # "CAP * ACK " is 10 bytes, leaving 500 bytes for the capabilities.
            # Sopel cannot allow multi-line requests, as it won't know how to
            # deal properly with multi-line ACK.
            # The spec says a client SHOULD send multiple requests; however
            # the spec also says that a server will ACK or NAK a whole request
            # at once. So technically, multiple REQs are not the same as a
            # single REQ.
            raise ValueError('Capability request too long: %s' % cap_req_text)

        self._cap_req: tuple[str, ...] = tuple(sorted(cap_req))
        self._handler: Optional[CapabilityHandler] = handler

    def __str__(self) -> str:
        caps = ", ".join(repr(cap) for cap in self._cap_req)
        handler = ""
        if self._handler and hasattr(self._handler, "__name__"):
            handler = " ({}())".format(self._handler.__name__)
        return "<capability {}{}>".format(caps, handler)

    @property
    def cap_req(self) -> tuple[str, ...]:
        """Capability request as a sorted tuple.

        This is the capability request that will be sent to the server as is.
        A request is acknowledged or denied for all the capabilities it
        contains, so the request ``(example/cap1, example/cap2)`` is not the
        same as two requests, one for ``example/cap1`` and the other for
        ``example/cap2``. This makes each request unique.
        """
        return self._cap_req

    def callback(
        self,
        bot: SopelWrapper,
        acknowledged: bool,
    ) -> tuple[bool, Optional[CapabilityNegotiation]]:
        """Execute the acknowlegement callback of a capability request.

        :param bot: a Sopel instance
        :param acknowledged: tell if the capability request is acknowledged
                             (``True``) or deny (``False``)
        :return: a 2-value tuple that contains if the request is done and the
                 result of the handler (if any)

        It executes the handler when the capability request receives an
        acknowledgement (either positive or negative), and returns the result.
        The handler's return value is used to know if the capability
        request is done, or if the bot must wait for resolution from the plugin
        that requested the capability.

        This method returns a 2-value tuple:

        * the first value tells if the negotiation is done for this request
        * the second is the handler's return value (if any)

        If no handler is registered, this automatically returns
        ``(True, None)``, as the negotiation is considered done (without any
        result).

        This doesn't prevent the handler from raising an exception.
        """
        result: Optional[CapabilityNegotiation] = None
        if self._handler is not None:
            result = self._handler(self.cap_req, bot, acknowledged)
            LOGGER.debug(
                'Cap request "%s" got "%s", '
                'executed successfuly with status: %s',
                ' '.join(self.cap_req),
                'ACK' if acknowledged else 'NAK',
                result.name,
            )
        return (result is None or result == CapabilityNegotiation.DONE, result)

    def __call__(
        self,
        handler: CapabilityHandler,
    ) -> capability:
        """Register a capability negotiation callback."""
        if self._handler is not None:
            raise RuntimeError("Cannot re-use capability decorator")
        self._handler = handler
        return self


def unblockable(
    function: Optional[Callable] = None,
) -> Callable:
    """Decorate a function to exempt it from Sopel's ignore system.

    For example, this can be used to ensure that important events such as
    ``JOIN`` are always recorded even if the user's nickname or hostname is
    :ref:`ignored <Ignoring Users>`::

        from sopel import plugin

        @plugin.event('JOIN')
        @plugin.unblockable
        def on_join_callable(bot, trigger):
            # do something when a user JOINs a channel
            # a blocked nickname or hostname *will* trigger this
            pass

    .. seealso::

        Sopel's :meth:`~sopel.bot.Sopel.dispatch` method.

    """
    def add_attribute(function):
        function.unblockable = True
        return function

    # hack to allow both @unblockable and @unblockable() to work
    if callable(function):
        return add_attribute(function)
    return add_attribute


def interval(*intervals: Union[int, float]) -> Callable:
    """Decorate a function to be called by the bot every *n* seconds.

    :param intervals: one or more duration(s), in seconds

    This decorator can be used multiple times for multiple intervals, or
    multiple intervals can be given in multiple arguments. The first time the
    function will be called is *n* seconds after the bot was started.

    Plugin functions decorated by ``interval`` must only take
    :class:`bot <sopel.bot.Sopel>` as their argument; they do not get a ``trigger``.
    The ``bot`` argument will not have a context, so functions like
    ``bot.say()`` will not have a default destination.

    There is no guarantee that the bot is connected to a server or in any
    channels when the function is called, so care must be taken.

    Example::

        from sopel import plugin

        @plugin.interval(5)
        def spam_every_5s(bot):
            if "#here" in bot.channels:
                bot.say("It has been five seconds!", "#here")

    """
    def add_attribute(function):
        function._sopel_callable = True
        if not hasattr(function, "interval"):
            function.interval = []
        for arg in intervals:
            if arg not in function.interval:
                function.interval.append(arg)
        return function

    return add_attribute


def rule(*patterns: Union[str, Pattern]) -> Callable:
    """Decorate a function to be called when a line matches the given pattern.

    :param patterns: one or more regular expression(s)

    Each argument is a regular expression which will trigger the function::

        @rule('hello', 'how')
            # will trigger once on "how are you?"
            # will trigger once on "hello, what's up?"

    This decorator can be used multiple times to add more rules::

        @rule('how')
        @rule('hello')
            # will trigger once on "how are you?"
            # will trigger once on "hello, what's up?"

    If the Sopel instance is in a channel, or sent a ``PRIVMSG``, where a
    string matching this expression is said, the function will execute. Note
    that captured groups here will be retrievable through the
    :class:`~sopel.trigger.Trigger` object later.

    Inside the regular expression, some special directives can be used.
    ``$nick`` will be replaced with the nick of the bot and ``,`` or ``:``, and
    ``$nickname`` will be replaced with the nick of the bot.

    .. versionchanged:: 7.0

        The :func:`rule` decorator can be called with multiple positional
        arguments, each used to add a rule. This is equivalent to decorating
        the same function multiple times with this decorator.

    .. note::

        The regex rule will match only once per line, starting at the beginning
        of the line only.

        To match for each time an expression is found, use the :func:`find`
        decorator instead. To match only once from anywhere in the line,
        use the :func:`search` decorator instead.

    """
    def add_attribute(function):
        function._sopel_callable = True
        if not hasattr(function, "rule"):
            function.rule = []
        for value in patterns:
            if value not in function.rule:
                function.rule.append(value)
        return function

    return add_attribute


def rule_lazy(*loaders: Callable) -> Callable:
    """Decorate a callable as a rule with lazy loading.

    :param loaders: one or more functions to generate a list of **compiled**
                    regexes to match URLs

    Each ``loader`` function must accept a ``settings`` parameter and return a
    list (or tuple) of **compiled** regular expressions::

        import re

        def loader(settings):
            return [re.compile(r'<your_rule_pattern>')]

    It will be called by Sopel when the bot parses the plugin to register rules
    to get its regexes. The ``settings`` argument will be the bot's
    :class:`sopel.config.Config` object.

    If any of the ``loader`` functions raises a
    :exc:`~sopel.plugins.exceptions.PluginError` exception, the rule will be
    ignored; it will not fail the plugin's loading.

    The decorated function will behave like any other :func:`callable`::

        from sopel import plugin

        @plugin.rule_lazy(loader)
        def my_rule_handler(bot, trigger):
            bot.say('Rule triggered by: %s' % trigger.group(0))

    .. versionadded:: 7.1

    .. seealso::

        When more than one loader is provided, they will be chained together
        with the :func:`sopel.tools.chain_loaders` function.

    """
    def decorator(function):
        function._sopel_callable = True
        if not hasattr(function, 'rule_lazy_loaders'):
            function.rule_lazy_loaders = []
        function.rule_lazy_loaders.extend(loaders)
        return function
    return decorator


def find(*patterns: Union[str, Pattern]) -> Callable:
    """Decorate a function to be called for each time a pattern is found in a line.

    :param patterns: one or more regular expression(s)

    Each argument is a regular expression which will trigger the function::

        @find('hello', 'here')
            # will trigger once on "hello you"
            # will trigger twice on "hello here"
            # will trigger once on "I'm right here!"

    This decorator can be used multiple times to add more rules::

        @find('here')
        @find('hello')
            # will trigger once on "hello you"
            # will trigger twice on "hello here"
            # will trigger once on "I'm right here!"

    If the Sopel instance is in a channel, or sent a ``PRIVMSG``, the function
    will execute for each time a received message matches an expression. Each
    match will also contain the position of the instance it found.

    Inside the regular expression, some special directives can be used.
    ``$nick`` will be replaced with the nick of the bot and ``,`` or ``:``, and
    ``$nickname`` will be replaced with the nick of the bot::

        @find('$nickname')
            # will trigger for each time the bot's nick is in a trigger

    .. versionadded:: 7.1

    .. note::

        The regex rule will match once for each non-overlapping match, from left
        to right, and the function will execute for each of these matches.

        To match only once from anywhere in the line, use the :func:`search`
        decorator instead. To match only once from the start of the line,
        use the :func:`rule` decorator instead.

    """
    def add_attribute(function):
        function._sopel_callable = True
        if not hasattr(function, "find_rules"):
            function.find_rules = []
        for value in patterns:
            if value not in function.find_rules:
                function.find_rules.append(value)
        return function

    return add_attribute


def find_lazy(*loaders: Callable) -> Callable:
    """Decorate a callable as a find rule with lazy loading.

    :param loaders: one or more functions to generate a list of **compiled**
                    regexes to match patterns in a line

    Each ``loader`` function must accept a ``settings`` parameter and return a
    list (or tuple) of **compiled** regular expressions::

        import re

        def loader(settings):
            return [re.compile(r'<your_rule_pattern>')]

    It will be called by Sopel when the bot parses the plugin to register the
    find rules to get its regexes. The ``settings`` argument will be the bot's
    :class:`sopel.config.Config` object.

    If any of the ``loader`` functions raises a
    :exc:`~sopel.plugins.exceptions.PluginError` exception, the find rule will
    be ignored; it will not fail the plugin's loading.

    The decorated function will behave like any other :func:`callable`::

        from sopel import plugin

        @plugin.find_lazy(loader)
        def my_find_rule_handler(bot, trigger):
            bot.say('Rule triggered by: %s' % trigger.group(0))

    .. versionadded:: 7.1

    .. seealso::

        When more than one loader is provided, they will be chained together
        with the :func:`sopel.tools.chain_loaders` function.

    """
    def decorator(function):
        function._sopel_callable = True
        if not hasattr(function, 'find_rules_lazy_loaders'):
            function.find_rules_lazy_loaders = []
        function.find_rules_lazy_loaders.extend(loaders)
        return function
    return decorator


def search(*patterns: Union[str, Pattern]) -> Callable:
    """Decorate a function to be called when a pattern matches anywhere in a line.

    :param patterns: one or more regular expression(s)

    Each argument is a regular expression which will trigger the function::

        @search('hello', 'here')
            # will trigger once on "hello you"
            # will trigger twice on "hello here"
            # will trigger once on "I'm right here!"

    This decorator can be used multiple times to add more search rules::

        @search('here')
        @search('hello')
            # will trigger once on "hello you"
            # will trigger twice on "hello here" (once per expression)
            # will trigger once on "I'm right here!"

    If the Sopel instance is in a channel, or sent a PRIVMSG, where a part
    of a string matching this expression is said, the function will execute.
    Note that captured groups here will be retrievable through the
    :class:`~sopel.trigger.Trigger` object later. The match will also contain
    the position of the first instance found.

    Inside the regular expression, some special directives can be used.
    ``$nick`` will be replaced with the nick of the bot and ``,`` or ``:``, and
    ``$nickname`` will be replaced with the nick of the bot::

        @search('$nickname')
            # will trigger once when the bot's nick is in a trigger

    .. versionadded:: 7.1

    .. note::

        The regex rule will match for the first instance only, starting from
        the left of the line, and the function will execute only once per
        regular expression.

        To match for each time an expression is found, use the :func:`find`
        decorator instead. To match only once from the start of the line,
        use the :func:`rule` decorator instead.

    """
    def add_attribute(function):
        function._sopel_callable = True
        if not hasattr(function, "search_rules"):
            function.search_rules = []
        for value in patterns:
            if value not in function.search_rules:
                function.search_rules.append(value)
        return function

    return add_attribute


def search_lazy(*loaders: Callable) -> Callable:
    """Decorate a callable as a search rule with lazy loading.

    :param loaders: one or more functions to generate a list of **compiled**
                    regexes to match patterns in a line

    Each ``loader`` function must accept a ``settings`` parameter and return a
    list (or tuple) of **compiled** regular expressions::

        import re

        def loader(settings):
            return [re.compile(r'<your_rule_pattern>')]

    It will be called by Sopel when the bot parses the plugin to register the
    search rules to get its regexes. The ``settings`` argument will be the
    bot's :class:`sopel.config.Config` object.

    If any of the ``loader`` functions raises a
    :exc:`~sopel.plugins.exceptions.PluginError` exception, the find rule will
    be ignored; it will not fail the plugin's loading.

    The decorated function will behave like any other :func:`callable`::

        from sopel import plugin

        @plugin.search_lazy(loader)
        def my_search_rule_handler(bot, trigger):
            bot.say('Rule triggered by: %s' % trigger.group(0))

    .. versionadded:: 7.1

    .. seealso::

        When more than one loader is provided, they will be chained together
        with the :func:`sopel.tools.chain_loaders` function.

    """
    def decorator(function):
        function._sopel_callable = True
        if not hasattr(function, 'search_rules_lazy_loaders'):
            function.search_rules_lazy_loaders = []
        function.search_rules_lazy_loaders.extend(loaders)
        return function
    return decorator


def thread(value: bool) -> Callable:
    """Decorate a function to specify if it should be run in a separate thread.

    :param value: if ``True``, the function is called in a separate thread;
                  otherwise, from the bot's main thread

    Functions run in a separate thread (as is the default) will not prevent the
    bot from executing other functions at the same time. Functions not run in a
    separate thread may be started while other functions are still running, but
    additional functions will not start until it is completed.
    """
    threaded = bool(value)

    def add_attribute(function):
        function.thread = threaded
        return function

    return add_attribute


def allow_bots(
    function: Optional[Callable] = None,
) -> Callable:
    """Decorate a function to specify that it should receive events from bots.

    On networks implementing the `Bot Mode specification`__, messages and
    other events from other clients that have identified themselves as bots
    will be tagged as such, and Sopel will ignore them by default. This
    decorator allows a function to opt into receiving these events.

    .. __: https://ircv3.net/specs/extensions/bot-mode
    """
    def add_attribute(function):
        function.allow_bots = True
        return function

    # hack to allow both @allow_bots and @allow_bots() to work
    if callable(function):
        return add_attribute(function)
    return add_attribute


def echo(
    function: Optional[Callable] = None,
) -> Callable:
    """Decorate a function to specify that it should receive echo messages.

    This decorator can be used to listen in on the messages that Sopel is
    sending and react accordingly.

    .. important::

        The decorated callable will receive *all* matching messages that Sopel
        sends, including output from the same callable. Take care to avoid
        creating feedback loops when using this feature.

    """
    def add_attribute(function):
        function.echo = True
        return function

    # hack to allow both @echo and @echo() to work
    if callable(function):
        return add_attribute(function)
    return add_attribute


def command(*command_list: str) -> Callable:
    """Decorate a function to set one or more commands that should trigger it.

    :param command_list: one or more command name(s) to match

    This decorator can be used to add multiple commands to one callable in a
    single line. The resulting match object will have the command as the first
    group; the rest of the line, excluding leading whitespace, as the second
    group; and parameters 1 through 4, separated by whitespace, as groups 3-6.

    Example::

        @command("hello")
            # If the command prefix is "\\.", this would trigger on lines
            # starting with ".hello".

        @command('j', 'join')
            # If the command prefix is "\\.", this would trigger on lines
            # starting with either ".j" or ".join".

    You can use a space in the command name to implement subcommands::

        @command('main sub1', 'main sub2')
            # For ".main sub1", trigger.group(1) will return "main sub1"
            # For ".main sub2", trigger.group(1) will return "main sub2"

    But in that case, be careful with the order of the names: if a more generic
    pattern is defined first, it will have priority over less generic patterns.
    So for instance, to have ``.main`` and ``.main sub`` working properly, you
    need to declare them like this::

        @command('main sub', 'main')
            # This command will react properly to ".main sub" and ".main"

    Then, you can check ``trigger.group(1)`` to know if it was used as
    ``main sub`` or just ``main`` in your callable. If you declare them in the
    wrong order, ``.main`` will have priority and you won't be able to take
    advantage of that.

    Another option is to declare command with subcommands only, like this::

        @command('main sub1')
            # this command will be triggered on .main sub1

        @command('main sub2')
            # this other command will be triggered on .main sub2

    In that case, ``.main`` won't trigger anything, and you won't have to
    inspect the trigger's groups to know which subcommand is triggered.

    .. note::

        If you use this decorator multiple times, remember that the decorators
        are invoked in the reverse order of appearance::

            # These two decorators...
            @command('hi')
            @command('hello')

            # ...are equivalent to this single decorator
            @command('hello', 'hi')

        See also the `Function Definitions`__ chapter from the Python
        documentation for more information about functions and decorators.

        .. __: https://docs.python.org/3/reference/compound_stmts.html#function-definitions

    .. note::

        The command name will be escaped for use in a regular expression.
        As such it is not possible to use something like ``.command\\d+`` to
        catch something like ``.command1`` or ``.command2``.

        You have several options at your disposal to replace a regex in the
        command name:

        * use a command alias
        * parse the arguments with your own regex within your plugin callable
        * use a :func:`rule` instead

    """
    def add_attribute(function):
        function._sopel_callable = True
        if not hasattr(function, "commands"):
            function.commands = []
        for command in command_list:
            if command not in function.commands:
                function.commands.append(command)
        return function
    return add_attribute


commands = command
"""Alias to :func:`command`."""


def nickname_command(*command_list: str) -> Callable:
    """Decorate a function to trigger on lines starting with "$nickname: command".

    :param command_list: one or more command name(s) to match

    This decorator can be used to add multiple commands to one callable in a
    single line. The resulting match object will have the command as the first
    group; the rest of the line, excluding leading whitespace, as the second
    group; and parameters 1 through 4, separated by whitespace, as groups 3-6.

    Example::

        @nickname_command("hello!")
            # Would trigger on "$nickname: hello!", "$nickname,   hello!",
            # "$nickname hello!", "$nickname hello! parameter1" and
            # "$nickname hello! p1 p2 p3 p4 p5 p6 p7 p8 p9".

    .. note::

        The command name will be escaped to be used in a regex command. As such
        it is not possible to use something like ``command\\d+`` to catch
        something like ``Bot: command1`` or ``Bot: command2``.

        You have several options at your disposal to replace a regex in the
        command name:

        * use a command alias
        * parse the arguments with your own regex within your plugin callable
        * use a :func:`rule`

        The :func:`rule` can be used with a ``$nick`` variable::

            @rule(r'$nick .*')
                # Would trigger on anything starting with "$nickname[:,]? ",
                # and would never have any additional parameters, as the
                # command would match the rest of the line.

    """
    def add_attribute(function):
        function._sopel_callable = True
        if not hasattr(function, 'nickname_commands'):
            function.nickname_commands = []
        for cmd in command_list:
            if cmd not in function.nickname_commands:
                function.nickname_commands.append(cmd)
        return function
    return add_attribute


nickname_commands = nickname_command
"""Alias to :func:`nickname_command`."""


def action_command(*command_list: str) -> Callable:
    """Decorate a function to trigger on CTCP ACTION lines.

    :param command_list: one or more command name(s) to match

    This decorator can be used to add multiple commands to one callable in a
    single line. The resulting match object will have the command as the first
    group; the rest of the line, excluding leading whitespace, as the second
    group; and parameters 1 through 4, separated by whitespace, as groups 3-6.

    Example::

        @action_command("hello!")
            # Would trigger on "/me hello!"

    .. versionadded:: 7.0

    .. note::

        The command name will be escaped for use in a regular expression.
        As such it is not possible to use something like ``/me command\\d+``
        to catch something like ``/me command1`` or ``/me command2``.

        You have several options at your disposal to replace a regex in the
        command name:

        * use a command alias
        * parse the arguments with your own regex within your plugin callable
        * use a :func:`rule`

        The :func:`rule` must be used with the :func:`ctcp` decorator::

            @rule(r'hello!?')
            @ctcp('ACTION')
                # Would trigger on "/me hello!" and "/me hello"

    """
    def add_attribute(function):
        function._sopel_callable = True
        if not hasattr(function, 'action_commands'):
            function.action_commands = []
        for cmd in command_list:
            if cmd not in function.action_commands:
                function.action_commands.append(cmd)
        return function
    return add_attribute


action_commands = action_command
"""Alias to :func:`action_command`."""


def label(value: str) -> Callable:
    """Decorate a function to add a rule label.

    :param value: a label for the rule

    The rule label allows the documentation and the logging system to refer
    to this function by its label. A function can have one and only one label::

        @label('on_hello')
        @rule('hello')
            # will trigger on hello, and be labelled as "on_hello"

    .. note::

        By default, the "label" of a callable will be its function name, which
        can be confusing for end-users: the goal of the ``label`` decorator is
        to make generic rules as user-friendly as commands are, by giving them
        some name that isn't tied to an identifier in the source code.

    """
    def add_attribute(function):
        function.rule_label = value
        return function
    return add_attribute


def priority(value: Literal['low', 'medium', 'high']) -> Callable:
    """Decorate a function to be executed with higher or lower priority.

    :param value: one of ``high``, ``medium``, or ``low``

    The priority allows you some control over the order of callable execution,
    if your plugin needs it. If a callable does not specify its ``priority``,
    Sopel assumes ``medium``.
    """
    def add_attribute(function):
        function.priority = value
        return function
    return add_attribute


def event(*event_list: str) -> Callable:
    """Decorate a function to be triggered on specific IRC events.

    :param event_list: one or more event name(s) on which to trigger

    This is one of a number of events, such as 'JOIN', 'PART', 'QUIT', etc.
    (More details can be found in RFC 1459.) When the Sopel bot is sent one of
    these events, the function will execute. Note that the default
    :meth:`rule` (``.*``) will match *any* line of the correct event type(s).
    If any rule is explicitly specified, it overrides the default.

    .. seealso::

        :class:`sopel.tools.events` provides human-readable names for many of the
        numeric events, which may help your code be clearer.

    """
    def add_attribute(function):
        function._sopel_callable = True
        if not hasattr(function, "event"):
            function.event = []
        for name in event_list:
            if name not in function.event:
                function.event.append(name)
        return function
    return add_attribute


def ctcp(
    function: Union[Callable, Optional[str]] = None,
    *command_list: str,
) -> Callable:
    """Decorate a callable to trigger on CTCP commands (mostly, ``ACTION``).

    :param command_list: one or more CTCP command(s) on which to trigger

    There are `various CTCP commands`__ to handle with this decorator, such as
    ``ACTION``, ``CLIENTINFO``, ``TIME``, and ``VERSION``::

        from sopel import plugin

        @plugin.ctcp('TIME')
        @plugin.rule('.*')
        def ctcp_time(bot, trigger):
            bot.say('Sorry, not a clock.')

    This decorator also works without parentheses, in which case it will trigger
    on CTCP ``ACTION``::

        from sopel import plugin

        @plugin.ctcp
        @plugin.rule('.*')
        def ctcp_action(bot, trigger):
            bot.reply('Why would you do that?!')

    .. versionadded:: 7.1

        This is now ``ctcp`` instead of ``intent``, and it can be called
        without argument, in which case it will assume ``ACTION``.

    .. note::

        This used to be ``@intent``, for a long dead feature in the IRCv3 spec.
        It is now replaced by ``@ctcp``, which can be used without arguments.
        In that case, Sopel will assume it should trigger on ``ACTION``.

        As ``sopel.module`` will be removed in Sopel 9, so will ``@intent``.

    .. __: https://datatracker.ietf.org/doc/html/draft-oakley-irc-ctcp-02#appendix-A
    """
    default_commands = ('ACTION',) + command_list
    if function is None:
        return ctcp(*default_commands)  # called as ``@ctcp()``
    elif callable(function):
        # called as ``@ctcp`` or ``@ctcp(function)``
        # or even ``@ctcp(function, 'ACTION', ...)``
        return ctcp(*default_commands)(function)

    # function is not None, and it is not a callable
    # called as ``@ctcp('ACTION', ...)``
    ctcp_commands = (function,) + command_list

    def add_attribute(function):
        function._sopel_callable = True
        if not hasattr(function, "ctcp"):
            function.ctcp = []
        for name in ctcp_commands:
            if name not in function.ctcp:
                function.ctcp.append(name)
        return function
    return add_attribute


def rate(
    user: int = 0,
    channel: int = 0,
    server: int = 0,
    *,
    message: Optional[str] = None,
) -> Callable:
    """Decorate a function to be rate-limited.

    :param user: seconds between permitted calls of this function by the same
                 user
    :param channel: seconds between permitted calls of this function in the
                    same channel, regardless of triggering user
    :param server: seconds between permitted calls of this function no matter
                   who triggered it or where
    :param message: optional keyword argument; default message sent as NOTICE
                    when a rate limit is reached

    How often a function can be triggered on a per-user basis, in a channel,
    or across the server (bot) can be controlled with this decorator. A value
    of ``0`` means no limit. If a function is given a rate of 20, that
    function may only be used once every 20 seconds in the scope corresponding
    to the parameter::

        from sopel import plugin

        @plugin.rate(10)
            # won't trigger if used more than once per 10s by a user

        @plugin.rate(10, 10)
            # won't trigger if used more than once per 10s by a user/channel

        @plugin.rate(10, 10, 2)
            # won't trigger if used more than once per 10s by a user/channel
            # and never more than once every 2s

    If a ``message`` is provided, it will be used as the default message sent
    as a ``NOTICE`` to the user who hit the rate limit::

        @rate(10, 10, 10, message='Hit the rate limit for this function.')
            # will send a NOTICE

    The message can contain placeholders which will be filled in:

    * ``nick``: the nick that hit the rate limit
    * ``channel``: the channel in which the rate limit was hit (will be
      ``'private-message'`` for private messages)
    * ``sender``: the sender (nick or channel) of the message which hit
      the rate limit
    * ``plugin``: the name of the plugin that hit the rate limit
    * ``label``: the label of the plugin handler that hit the rate limit
    * ``time_left``: the time remaining before the rate limit expires, as
      a string
    * ``time_left_sec``: the time remaining before the rate limit expires,
      expressed in number of seconds
    * ``rate_limit``: the rate limit, as a string
    * ``rate_limit_sec``: the rate limit, expressed in number of seconds
    * ``rate_limit_type``: the type of rate limit that was hit (one of
      ``user, group, global``)

    For example::

        @rate(10, 10, 2, message='Sorry {nick}, you hit the {rate_limit_type} rate limit!')

    .. versionchanged:: 8.0

        Optional keyword argument ``message`` was added in Sopel 8.

    .. note::

        Users on the admin list in Sopel's configuration are exempted from rate
        limits.

    .. seealso::

        You can control each rate limit separately, with their own custom
        message using :func:`rate_user`, :func:`rate_channel`, or
        :func:`rate_global`.

    """
    def add_attribute(function):
        if not hasattr(function, 'user_rate'):
            function.user_rate = user
        if not hasattr(function, 'channel_rate'):
            function.channel_rate = channel
        if not hasattr(function, 'global_rate'):
            function.global_rate = server
        function.default_rate_message = message
        return function
    return add_attribute


def rate_user(
    rate: int,
    message: Optional[str] = None,
) -> Callable:
    """Decorate a function to be rate-limited for a user.

    :param rate: seconds between permitted calls of this function by the same
                 user
    :param message: optional; message sent as NOTICE when a user hits the limit

    This decorator can be used alone or with the :func:`rate` decorator, as it
    will always take precedence::

        @rate(10, 10, 10)
        @rate_user(20, 'You hit your rate limit for this function.')
            # user limit will be set to 20, other to 10
            # will send a NOTICE only when a user hits their own limit
            # as other rate limits don't have any message set

    The message can contain the same placeholders supported by :func:`rate`::

        @rate_user(5, 'Sorry {nick}, you hit your {rate_limit_sec}s limit!')

    If you don't provide a message, the default message set by :func:`rate`
    (if any) will be used instead.

    .. versionadded:: 8.0

    .. note::

        Users on the admin list in Sopel's configuration are exempted from rate
        limits.

    """
    def add_attribute(function):
        function.user_rate = rate
        function.user_rate_message = message
        return function
    return add_attribute


def rate_channel(
    rate: int,
    message: Optional[str] = None,
) -> Callable:
    """Decorate a function to be rate-limited for a channel.

    :param rate: seconds between permitted calls of this function in the same
                 channel, regardless of triggering user
    :param message: optional; message sent as NOTICE when a user hits the limit

    This decorator can be used alone or with the :func:`rate` decorator, as it
    will always take precedence::

        @rate(10, 10, 10)
        @rate_channel(5, 'You hit the channel rate limit for this function.')
            # channel limit will be set to 5, other to 10
            # will send a NOTICE only when a user hits the channel limit
            # as other rate limits don't have any message set

    If you don't provide a message, the default message set by :func:`rate`
    (if any) will be used instead.

    The message can contain the same placeholders supported by :func:`rate`::

        @rate_channel(
            5,
            'Sorry {nick}, you hit the {rate_limit_sec}s limit for the {channel} channel!',
        )

    .. versionadded:: 8.0

    .. note::

        Users on the admin list in Sopel's configuration are exempted from rate
        limits.

    """
    def add_attribute(function):
        function.channel_rate = rate
        function.channel_rate_message = message
        return function
    return add_attribute


def rate_global(
    rate: int,
    message: Optional[str] = None,
) -> Callable:
    """Decorate a function to be rate-limited for the whole server.

    :param rate: seconds between permitted calls of this function no matter who
                 triggered it or where
    :param message: optional; message sent as NOTICE when a user hits the limit

    This decorator can be used alone or with the :func:`rate` decorator, as it
    will always take precedence.

    For example::

        @rate(10, 10, 10)
        @rate_global(5, 'You hit the global rate limit for this function.')
            # global limit will be set to 5, other to 10
            # will send a NOTICE only when a user hits the global limit
            # as other rate limits don't have any message set

    If you don't provide a message, the default message set by :func:`rate`
    (if any) will be used instead.

    The message can contain the same placeholders supported by :func:`rate`::

        @rate_global(5, 'Sorry {nick}, you hit the 5s limit!')

    .. versionadded:: 8.0

    .. note::

        Users on the admin list in Sopel's configuration are exempted from rate
        limits.

    """
    def add_attribute(function):
        function.global_rate = rate
        function.global_rate_message = message
        return function
    return add_attribute


def require_privmsg(
    message: Union[Callable, Optional[str]] = None,
    reply: bool = False,
) -> Callable:
    """Decorate a function to only be triggerable from a private message.

    :param message: optional message said if triggered in a channel
    :param reply: use :meth:`~sopel.bot.Sopel.reply` instead of
                  :meth:`~sopel.bot.Sopel.say` when ``True``; defaults to
                  ``False``

    A decorated plugin callable will be triggered only by messages sent to the
    bot in private::

        from sopel import plugin

        @plugin.command('.shh')
        @plugin.require_privmsg('PM only command.')
        def confidential_command(bot, trigger):
            # trigger on private messages only

    If the decorated function is triggered by a channel message, the ``message``
    will be said if given. By default, it uses :meth:`bot.say()
    <.bot.Sopel.say>`, but when ``reply`` is ``True``, then it uses
    :meth:`bot.reply() <.bot.Sopel.reply>` instead.

    This decorator also works without parentheses, if you want its default (no
    arguments) behavior::

        from sopel import plugin

        @plugin.command('.shh')
        @plugin.require_privmsg
        def confidential_command(bot, trigger):
            # trigger on private messages only

    .. versionchanged:: 7.0
        Added the ``reply`` parameter.
    """
    def actual_decorator(function):
        @functools.wraps(function)
        def guarded(bot, trigger, *args, **kwargs):
            if trigger.is_privmsg:
                return function(bot, trigger, *args, **kwargs)
            else:
                if message and not callable(message):
                    if reply:
                        bot.reply(message)
                    else:
                        bot.say(message)
        return guarded

    # Hack to allow decorator without parens
    if callable(message):
        return actual_decorator(message)
    return actual_decorator


def require_chanmsg(
    message: Union[Callable, Optional[str]] = None,
    reply: bool = False,
) -> Callable:
    """Decorate a function to only be triggerable from a channel message.

    :param message: optional message said if triggered in private message
    :param reply: use :meth:`~.bot.Sopel.reply` instead of
                  :meth:`~.bot.Sopel.say` when ``True``; defaults to ``False``

    A decorated plugin callable will be triggered only by messages from a
    channel::

        from sopel import plugin

        @plugin.command('.mtopic')
        @plugin.require_chanmsg('Channel only command.')
        def manage_topic(bot, trigger):
            # trigger on channel messages only

    If the decorated function is triggered by a private message, the ``message``
    will be said if given. By default, it uses :meth:`bot.say()
    <.bot.Sopel.say>`, but when ``reply`` is ``True`` it uses
    :meth:`bot.reply() <.bot.Sopel.reply>` instead.

    This decorator also works without parentheses, if you want its default (no
    arguments) behavior::

        from sopel import plugin

        @plugin.command('.mtopic')
        @plugin.require_chanmsg
        def manage_topic(bot, trigger):
            # trigger on channel messages only

    .. versionchanged:: 7.0
        Added the ``reply`` parameter.
    """
    def actual_decorator(function):
        @functools.wraps(function)
        def guarded(bot, trigger, *args, **kwargs):
            if not trigger.is_privmsg:
                return function(bot, trigger, *args, **kwargs)
            else:
                if message and not callable(message):
                    if reply:
                        bot.reply(message)
                    else:
                        bot.say(message)
        return guarded

    # Hack to allow decorator without parens
    if callable(message):
        return actual_decorator(message)
    return actual_decorator


def require_account(
    message: Union[Callable, Optional[str]] = None,
    reply: bool = False,
) -> Callable:
    """Decorate a function to require services/NickServ authentication.

    :param message: optional message to say if a user without authentication
                    tries to trigger this function
    :param reply: use :meth:`~.bot.Sopel.reply` instead of
                  :meth:`~.bot.Sopel.say` when ``True``; defaults to ``False``

    A decorated plugin callable will be triggered only if the triggering user is
    logged into a network services account::

        from sopel import plugin

        @plugin.command('.regonly')
        @plugin.require_account('Registered users only.')
        def logged_in_command(bot, trigger):
            # trigger only if user is logged in to services

    If an unauthenticated user triggers the decorated function, the ``message``
    will be said if given. By default, it uses :meth:`bot.say()
    <.bot.Sopel.say>`, but when ``reply`` is ``True`` it uses :meth:`bot.reply()
    <.bot.Sopel.reply>` instead.

    This decorator also works without parentheses, if you want its default (no
    arguments) behavior::

        from sopel import plugin

        @plugin.command('.regonly')
        @plugin.require_account
        def logged_in_command(bot, trigger):
            # trigger only if user is logged in to services

    .. versionadded:: 7.0
    .. note::

        Only some networks support services authentication, and not all of
        those implement the standards required for clients like Sopel to
        determine authentication status. This decorator will block *all* use
        of functions it decorates on networks that lack the relevant features.

    .. seealso::

        The value of the :class:`trigger<.trigger.Trigger>`'s
        :attr:`account <.trigger.Trigger.account>` property determines whether
        this requirement is satisfied, and the property's documentation
        includes up-to-date details on what features a network must
        support to allow Sopel to fetch account information.

    """
    def actual_decorator(function):
        @functools.wraps(function)
        def guarded(bot, trigger, *args, **kwargs):
            if not trigger.account:
                if message and not callable(message):
                    if reply:
                        bot.reply(message)
                    else:
                        bot.say(message)
            else:
                return function(bot, trigger, *args, **kwargs)
        return guarded

    # Hack to allow decorator without parens
    if callable(message):
        return actual_decorator(message)

    return actual_decorator


def require_privilege(
    level: AccessLevel,
    message: Optional[str] = None,
    reply: bool = False,
) -> Callable:
    """Decorate a function to require at least the given channel permission.

    :param level: required privilege level to use this command
    :param message: optional message said to insufficiently privileged user
    :param reply: use :meth:`~.bot.Sopel.reply` instead of
                  :meth:`~.bot.Sopel.say` when ``True``; defaults to ``False``

    ``level`` can be one of the privilege level constants defined in this
    module. If the user does not have at least that privilege, the bot will say
    ``message`` if given. By default, it uses :meth:`bot.say()
    <.bot.Sopel.say>`, but when ``reply`` is ``True`` it uses :meth:`bot.reply()
    <.bot.Sopel.reply>` instead.

    Use of ``require_privilege()`` implies :func:`require_chanmsg`.

    .. versionchanged:: 7.0
        Added the ``reply`` parameter.

    .. versionchanged:: 8.0
        Decorated callables no longer run in response to private messages.
    """
    def actual_decorator(function):
        @functools.wraps(function)
        def guarded(bot, trigger, *args, **kwargs):
            # If this is a privmsg, do not trigger
            if trigger.is_privmsg:
                return
            channel_privs = bot.channels[trigger.sender].privileges
            allowed = channel_privs.get(trigger.nick, 0) >= level
            if not trigger.is_privmsg and not allowed:
                if message and not callable(message):
                    if reply:
                        bot.reply(message)
                    else:
                        bot.say(message)
            else:
                return function(bot, trigger, *args, **kwargs)
        return guarded
    return actual_decorator


def require_admin(
    message: Union[Callable, Optional[str]] = None,
    reply: bool = False,
) -> Callable:
    """Decorate a function to require the triggering user to be a bot admin.

    :param message: optional message said to non-admin user
    :param reply: use :meth:`~.bot.Sopel.reply` instead of
                  :meth:`~.bot.Sopel.say` when ``True``; defaults to ``False``

    A decorated plugin callable will be triggered only if the triggering user is
    an admin of the bot, according to its configuration::

        from sopel import plugin

        @plugin.command('.adminonly')
        @plugin.require_admin('Bot admin only command.')
        def admin_command(bot, trigger):
            # trigger only if user is a bot admin

    The bot's :attr:`~.config.core_section.CoreSection.owner` is also an admin.

    When the triggering user is not an admin, the command is not run, and the
    bot will say the ``message`` if given. By default, it uses :meth:`bot.say()
    <.bot.Sopel.say>`, but when ``reply`` is ``True`` it uses :meth:`bot.reply()
    <.bot.Sopel.reply>` instead.

    This decorator also works without parentheses, if you want its default (no
    arguments) behavior::

        from sopel import plugin

        @plugin.command('.adminonly')
        @plugin.require_admin
        def admin_command(bot, trigger):
            # trigger only if user is a bot admin

    .. versionchanged:: 7.0
        Added the ``reply`` parameter.
    """
    def actual_decorator(function):
        @functools.wraps(function)
        def guarded(bot, trigger, *args, **kwargs):
            if not trigger.admin:
                if message and not callable(message):
                    if reply:
                        bot.reply(message)
                    else:
                        bot.say(message)
            else:
                return function(bot, trigger, *args, **kwargs)
        return guarded

    # Hack to allow decorator without parens
    if callable(message):
        return actual_decorator(message)

    return actual_decorator


def require_owner(
    message: Union[Callable, Optional[str]] = None,
    reply: bool = False,
) -> Callable:
    """Decorate a function to require the triggering user to be the bot owner.

    :param message: optional message said to non-owner user
    :param reply: use :meth:`~.bot.Sopel.reply` instead of
                  :meth:`~.bot.Sopel.say` when ``True``; defaults to ``False``

    A decorated plugin callable will be triggered only if the triggering user is
    recognized as the bot's owner, according to its configuration::

        from sopel import plugin

        @plugin.command('.owneronly')
        @plugin.require_owner('Bot owner only command.')
        def owner_command(bot, trigger):
            # trigger only if user is the bot's owner

    When the triggering user is not the bot's owner, the command is not run,
    and the bot will say ``message`` if given. By default, it uses
    :meth:`bot.say() <.bot.Sopel.say>`, but when ``reply`` is ``True`` it uses
    :meth:`bot.reply() <.bot.Sopel.reply>` instead.

    This decorator also works without parentheses, if you want its default (no
    arguments) behavior::

        from sopel import plugin

        @plugin.command('.owneronly')
        @plugin.require_owner
        def owner_command(bot, trigger):
            # trigger only if user is the bot's owner

    .. versionchanged:: 7.0
        Added the ``reply`` parameter.
    """
    def actual_decorator(function):
        @functools.wraps(function)
        def guarded(bot, trigger, *args, **kwargs):
            if not trigger.owner:
                if message and not callable(message):
                    if reply:
                        bot.reply(message)
                    else:
                        bot.say(message)
            else:
                return function(bot, trigger, *args, **kwargs)
        return guarded

    # Hack to allow decorator without parens
    if callable(message):
        return actual_decorator(message)
    return actual_decorator


def require_bot_privilege(
    level: AccessLevel,
    message: Optional[str] = None,
    reply: bool = False,
) -> Callable:
    """Decorate a function to require a minimum channel privilege for the bot.

    :param level: minimum channel privilege the bot needs for this function
    :param message: optional message said if the bot's channel privilege level
                    is insufficient
    :param reply: use :meth:`~.bot.Sopel.reply` instead of
                  :meth:`~.bot.Sopel.say` when ``True``; defaults to ``False``

    ``level`` can be one of the privilege level constants defined in this
    module. If the bot does not have the privilege, the bot will say
    ``message`` if given. By default, it uses :meth:`bot.say()
    <.bot.Sopel.say>`, but when ``reply`` is ``True`` it uses :meth:`bot.reply()
    <.bot.Sopel.reply>` instead.

    Use of ``require_bot_privilege()`` implies :func:`require_chanmsg`.

    .. versionadded:: 7.1
    .. versionchanged:: 8.0
        Decorated callables no longer run in response to private messages.
    """
    def actual_decorator(function):
        @functools.wraps(function)
        def guarded(bot, trigger, *args, **kwargs):
            # If this is a privmsg, do not trigger
            if trigger.is_privmsg:
                return

            if not bot.has_channel_privilege(trigger.sender, level):
                if message and not callable(message):
                    if reply:
                        bot.reply(message)
                    else:
                        bot.say(message)
            else:
                return function(bot, trigger, *args, **kwargs)
        return guarded
    return actual_decorator


def url(*url_rules: str) -> Callable:
    """Decorate a function to handle URLs.

    :param url_rules: one or more regex pattern(s) to match URLs

    This decorator takes a regex string that will be matched against URLs in a
    message. The function it decorates is like any other callable::

        from sopel import plugin

        @plugin.url(r'https://example.com/bugs/([a-z0-9]+)')
        @plugin.url(r'https://short.com/([a-z0-9]+)')
        def handle_example_bugs(bot, trigger):
            bot.reply('Found bug ID #%s' % trigger.group(1))

    The ``bot`` is an instance of :class:`~sopel.bot.SopelWrapper`, and
    ``trigger`` is the usual :class:`~sopel.trigger.Trigger` object.

    Under the hood, when Sopel collects the decorated handler it uses an
    instance of :class:`sopel.plugins.rules.URLCallback` to register it to its
    :attr:`rules manager <sopel.bot.Sopel.rules>` and its
    :meth:`~sopel.plugins.rules.Manager.register_url_callback` method.

    .. versionchanged:: 7.0

        The same function can be decorated multiple times with :func:`url`
        to register different URL patterns.

    .. versionchanged:: 7.0

        More than one pattern can be provided as positional argument at once.

    .. versionchanged:: 7.1

        The ``match`` parameter is obsolete and can be omitted. When present
        however, it represents the same match as the ``trigger`` argument.

        This behavior will be kept for backward compatibility and will be
        removed in Sopel 9.

    .. seealso::

        To detect URLs, Sopel uses a matching pattern built from a list of URL
        schemes, configured by
        :attr:`~sopel.config.core_section.CoreSection.auto_url_schemes`.

    """
    def actual_decorator(function):
        function._sopel_callable = True
        if not hasattr(function, 'url_regex'):
            function.url_regex = []
        for url_rule in url_rules:
            url_regex = re.compile(url_rule)
            if url_regex not in function.url_regex:
                function.url_regex.append(url_regex)
        return function
    return actual_decorator


def url_lazy(*loaders: Callable) -> Callable:
    """Decorate a function to handle URL, using lazy-loading for its regex.

    :param loaders: one or more functions to generate a list of **compiled**
                    regexes to match URLs.

    Each ``loader`` function must accept a ``settings`` parameter and return a
    list (or tuple) of **compiled** regular expressions::

        import re

        def loader(settings):
            return [re.compile(r'<your_url_pattern>')]

    It will be called by Sopel when the bot parses the plugin to register URL
    callbacks to get its regexes. The ``settings`` argument will be the bot's
    :class:`sopel.config.Config` object.

    If any of the ``loader`` functions raises a
    :exc:`~sopel.plugins.exceptions.PluginError` exception, the URL callback
    will be ignored; it will not fail the plugin's loading.

    The decorated function will behave like any other :func:`callable`::

        from sopel import plugin

        @plugin.url_lazy(loader)
        def my_url_handler(bot, trigger):
            bot.say('URL found: %s' % trigger.group(0))

    .. versionadded:: 7.1
    .. seealso::

        When more than one loader is provided, they will be chained together
        with the :func:`sopel.tools.chain_loaders` function.

    """
    def decorator(function):
        function._sopel_callable = True
        if not hasattr(function, 'url_lazy_loaders'):
            function.url_lazy_loaders = []
        function.url_lazy_loaders.extend(loaders)
        return function
    return decorator


class example:
    """Decorate a function with an example, and optionally test output.

    :param msg: the example command (required; see below)
    :param result: the command's expected output (optional; see below)
    :param privmsg: if ``True``, the example will be tested as if it was
                    received in a private message to the bot; otherwise,
                    in a channel (optional; default ``False``)
    :param admin: whether to treat the test message as having come from a
                  bot admin (optional; default ``False``)
    :param owner: whether to treat the test message as having come from
                  the bot's owner (optional; default ``False``)
    :param repeat: how many times to repeat the test; useful for commands
                   that return random results (optional; default ``1``)
    :param re: if ``True``, the ``result`` is interpreted as a regular
               expression and used to match the command's output
               (optional; see below)
    :param ignore: list of regular expression patterns to match ignored
                   output (optional; see below)
    :param user_help: whether this example should be included in
                      user-facing help output such as `.help command`
                      (optional; default ``False``; see below)
    :param online: if ``True``, |pytest|_ will mark this example as "online"
                   (optional; default ``False``; see below)
    :param vcr: if ``True``, this example's HTTP requests & responses will
                be recorded for later reuse (optional; default ``False``;
                see below)

    .. |pytest| replace:: ``pytest``
    .. _pytest: https://pypi.org/project/pytest/

    For compatibility with the built-in help plugin, ``msg`` must use the
    default :attr:`~sopel.config.core_section.CoreSection.help_prefix` if it
    is a prefixed command. Other command types should give example invocations
    that work with Sopel's default settings, especially if using the "example
    test" functionality to automatically generate a test(s) for the function.

    The presence of a ``result`` will generate tests automatically when Sopel's
    test suite is run, using ``msg`` as input. The exact behavior of the tests
    depends on the remaining optional ``example`` arguments.

    Passing ``re=True``, in particular, is useful for matching ``result``\\s
    that are random and/or dependent on an external API. This way, an example
    test can check the format of the result without caring about the exact data.

    Giving a list of ``ignore``d patterns is helpful for commands that may
    return intermittent errors (mostly calls to an external API that isn't
    necessarily stable), especially when coupled with the ``repeat`` parameter.

    By default, Sopel's help plugin will display only one example (the one
    closest to the function's `def` statement, due to how decorators work). You
    can override this choice or include multiple examples by passing
    ``user_help=True`` to one or more ``example`` decorator(s).

    Passing ``online=True`` makes that particular example skippable if Sopel's
    test suite is run in offline mode, which is mostly useful to make life
    easier for other developers working on Sopel without Internet access.

    Finally, ``vcr=True`` records the example's HTTP requests and responses for
    replaying during later test runs. It can be an alternative (or complement)
    to ``online``, and is especially useful for testing plugin code that calls
    on inconsistent or flaky remote APIs. The recorded "cassettes" of responses
    can be committed alongside the code for use by CI services, etc. (See
    `VCR.py <https://github.com/kevin1024/vcrpy>`_ & `pytest-vcr
    <https://github.com/ktosiek/pytest-vcr>`_)
    """
    def __init__(
        self,
        msg: str,
        result: Optional[Union[str, Iterable[str]]] = None,
        privmsg: bool = False,
        admin: bool = False,
        owner: bool = False,
        repeat: int = 1,
        re: bool = False,
        ignore: Optional[Union[str, Iterable[str]]] = None,
        user_help: bool = False,
        online: bool = False,
        vcr: bool = False,
    ):
        # Wrap result into a list for get_example_test
        self.result: Optional[list[str]] = None
        if isinstance(result, str):
            self.result = [result]
        elif result is not None:
            self.result = list(result)

        self.use_re = re
        self.msg = msg
        self.privmsg = privmsg
        self.admin = admin
        self.owner = owner
        self.repeat = repeat
        self.online = online
        self.vcr = vcr

        self.ignore: list[str] = []
        if isinstance(ignore, str):
            self.ignore = [ignore]
        elif ignore is not None:
            self.ignore = list(ignore)

        self.user_help = user_help

    def __call__(self, func):
        # mypy (as of v1.4) doesn't recognize the below check as adding an
        # "example" attribute to `func` if it's missing, so the `func`
        # argument isn't typed yet.
        # When we're ready to type-hint `loader`, we can make a TypeVar like
        # `Callable[[SopelWrapper, Trigger], Any]` (but with the attributes
        # the loader adds) for use in places like this.
        if not hasattr(func, "example"):
            func.example = []

        import sys

        # only inject test-related stuff if we're running tests
        # see https://stackoverflow.com/a/44595269/5991
        if 'pytest' in sys.modules and self.result:
            from sopel.tests import pytest_plugin

            # avoids doing `import pytest` and causing errors when
            # dev-dependencies aren't installed
            pytest = sys.modules['pytest']

            test = pytest_plugin.get_example_test(
                func, self.msg, self.result, self.privmsg, self.admin,
                self.owner, self.repeat, self.use_re, self.ignore
            )

            if self.online:
                test = pytest.mark.online(test)

            if self.vcr:
                test = pytest.mark.vcr(test)

            pytest_plugin.insert_into_module(
                test, func.__module__, func.__name__, 'test_example'
            )
            pytest_plugin.insert_into_module(
                pytest_plugin.get_disable_setup(),
                func.__module__,
                func.__name__,
                'disable_setup',
            )

        record = {
            "example": self.msg,
            "result": self.result,
            # flags
            "is_private_message": self.privmsg,
            "is_help": self.user_help,
            "is_pattern": self.use_re,
            "is_admin": self.admin,
            "is_owner": self.owner,
        }
        func.example.append(record)
        return func


def output_prefix(prefix: str) -> Callable:
    """Decorate a function to add a prefix on its output.

    :param prefix: the prefix to add (must include trailing whitespace if
                   desired; Sopel does not assume it should add anything)

    Prefix will be added to text sent through:

    * :meth:`bot.say <sopel.bot.SopelWrapper.say>`
    * :meth:`bot.notice <sopel.bot.SopelWrapper.notice>`

    """
    def add_attribute(function):
        function.output_prefix = prefix
        return function
    return add_attribute



"""
isup.py - Sopel Website Status Check Plugin
Copyright 2011, Elsie Powell http://embolalia.com
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

import requests

from sopel import plugin


PLUGIN_OUTPUT_PREFIX = '[isup] '


def get_site_url(site):
    """Get a ``site`` URL

    :param str site: the site to get URL for
    :return: a valid site URL
    :raise ValueError: when site is empty, or isn't well formatted

    The ``site`` argument is checked: its scheme must be ``http`` or ``https``,
    or a :exc:`ValueError` is raised.

    If the ``site`` does not have a scheme, ``http`` is used. If it doesn't
    have a TLD, a :exc:`ValueError` is raised.
    """
    site = site.strip() if site else ''
    if not site:
        raise ValueError('What site do you want to check?')

    if not site.startswith(('http://', 'https://')):
        if '://' in site:
            protocol = site.split('://')[0] + '://'
            raise ValueError('Try it again without the %s' % protocol)

        site = 'http://' + site

    domain = site.split('/')[2].split(':')[0]
    if '.' not in domain:
        raise ValueError('I need a fully qualified domain name (with a dot).')
    if domain.endswith(('.local', '.example', '.test', '.invalid', '.localhost')):
        raise ValueError("I can't check LAN-local or invalid domains.")

    return site


def handle_isup(bot, trigger, secure=True):
    """Handle the ``bot`` command from ``trigger``

    :param bot: Sopel instance
    :type bot: :class:`sopel.bot.SopelWrapper`
    :param trigger: Command's trigger instance
    :type trigger: :class:`sopel.trigger.Trigger`
    :param bool secure: Check SSL error if ``True`` (the default)
    """
    try:
        site = get_site_url(trigger.group(2))
    except ValueError as error:
        bot.reply(str(error))
        return

    try:
        response = requests.head(site, verify=secure, timeout=(10.0, 5.0))
        response.raise_for_status()
    except requests.exceptions.SSLError:
        bot.say(
            '{} looks down to me (SSL error). Try using `{}isupinsecure`.'
            .format(site, bot.config.core.help_prefix))
    except requests.HTTPError:
        bot.say(
            '{} looks down to me (HTTP {} "{}").'
            .format(site, response.status_code, response.reason))
    except requests.ConnectTimeout:
        bot.say(
            '{} looks down to me (timed out while connecting).'
            .format(site))
    except requests.ReadTimeout:
        bot.say(
            '{} looks down to me (timed out waiting for reply).'
            .format(site))
    except requests.ConnectionError:
        bot.say(
            '{} looks down to me (connection error).'
            .format(site))
    except ValueError:
        bot.reply('"{}" is not a valid URL.'.format(site))
    else:
        # If no exception happened, the request must have succeeded.
        bot.say(site + ' looks fine to me.')


@plugin.command('isupinsecure')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def isup_insecure(bot, trigger):
    """Check if a website is up (without verifying HTTPS)."""
    handle_isup(bot, trigger, secure=False)


@plugin.command('isup')
@plugin.example('.isup google.com')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def isup(bot, trigger):
    """Check if a website is up or not."""
    handle_isup(bot, trigger, secure=True)



from __future__ import annotations

import logging
from logging.config import dictConfig
import os

from sopel import tools
from sopel.lifecycle import deprecated


class IrcLoggingHandler(logging.Handler):
    """Logging handler for output to an IRC channel.

    :param bot: a Sopel instance
    :type bot: :class:`sopel.bot.Sopel`
    :param level: minimum level of log messages to report through this handler
    :type level: :ref:`logging level <levels>`

    Implementation of a :class:`logging.Handler`.
    """
    def __init__(self, bot, level):
        super().__init__(level)
        self._bot = bot
        self._channel = bot.settings.core.logging_channel

    def emit(self, record):
        """Emit a log ``record`` to the IRC channel.

        :param record: the log record to output
        :type record: :class:`logging.LogRecord`
        """
        if not self._bot.connection_registered:
            # Don't emit logs to IRC when the bot is not connected & registered.
            return

        try:
            msg = self.format(record).replace('\n', ' ')
            self._bot.say(msg, self._channel)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:  # TODO: Be specific
            self.handleError(record)


class ChannelOutputFormatter(logging.Formatter):
    """Output formatter for log messages destined for an IRC channel.

    :param fmt: log line format
    :type fmt: :ref:`format string <formatstrings>`
    :param datefmt: date format
    :type datefmt: :ref:`format string <formatstrings>`

    Implementation of a :class:`logging.Formatter`.
    """
    def __init__(self, fmt='[%(filename)s] %(message)s', datefmt=None):
        super().__init__(fmt=fmt, datefmt=datefmt)

    def formatException(self, exc_info):
        """Format the exception info as a string for output.

        :param tuple exc_info: standard exception information returned by
                               :func:`~sys.exc_info`
        """
        # logging will throw a newline between the message and this, but
        # that's fine because Sopel will strip it back out anyway
        return ' - ' + repr(exc_info[1])


def setup_logging(settings):
    """Set up logging based on the bot's configuration ``settings``.

    :param settings: configuration settings object
    :type settings: :class:`sopel.config.Config`
    """
    log_directory = settings.core.logdir
    base_level = settings.core.logging_level or 'WARNING'
    base_format = settings.core.logging_format
    base_datefmt = settings.core.logging_datefmt

    logging_config = {
        'version': 1,
        'formatters': {
            'sopel': {
                'format': base_format,
                'datefmt': base_datefmt,
            },
            'raw': {
                'format': '%(asctime)s %(message)s',
                'datefmt': base_datefmt,
            },
        },
        'loggers': {
            # all purpose, sopel root logger
            'sopel': {
                'level': base_level,
                'handlers': ['console', 'logfile', 'errorfile'],
            },
            # raw IRC log
            'sopel.raw': {
                'level': 'DEBUG',
                'propagate': False,
                'handlers': ['raw'],
            },
            # asynchat exception logger
            'sopel.exceptions': {
                'level': 'INFO',
                'propagate': False,
                'handlers': ['exceptionfile'],
            },
            # asyncio logging
            'asyncio': {
                'level': 'DEBUG',
                'handlers': ['console'],
            },
        },
        'handlers': {
            # output on stderr
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'sopel',
            },
            # generic purpose log file
            'logfile': {
                'level': 'DEBUG',
                'class': 'logging.handlers.TimedRotatingFileHandler',
                'filename': os.path.join(
                    log_directory, settings.basename + '.sopel.log'),
                'when': 'midnight',
                'formatter': 'sopel',
            },
            # caught error log file
            'errorfile': {
                'level': 'ERROR',
                'class': 'logging.handlers.TimedRotatingFileHandler',
                'filename': os.path.join(
                    log_directory, settings.basename + '.error.log'),
                'when': 'midnight',
                'formatter': 'sopel',
            },
            # uncaught error file
            'exceptionfile': {
                'level': 'ERROR',
                'class': 'logging.handlers.TimedRotatingFileHandler',
                'filename': os.path.join(
                    log_directory, settings.basename + '.exceptions.log'),
                'when': 'midnight',
                'formatter': 'sopel',
            },
            # raw IRC log file
            'raw': {
                'level': 'DEBUG',
                'class': 'logging.handlers.TimedRotatingFileHandler',
                'filename': os.path.join(
                    log_directory, settings.basename + '.raw.log'),
                'when': 'midnight',
                'formatter': 'raw',
            },
        },
    }
    dictConfig(logging_config)


@deprecated(
    reason='use sopel.tools.get_logger instead',
    version='7.0',
    warning_in='8.0',
    removed_in='9.0',
)
def get_logger(name=None):
    """Return a logger for a module, if the name is given.

    .. deprecated:: 7.0

        Sopel's own code should use :func:`logging.getLogger(__name__)
        <logging.getLogger>` instead, and external plugins should use
        :func:`sopel.tools.get_logger`.

        This will emit a deprecation warning in Sopel 8.0, and it will be
        removed in Sopel 9.0.

    """
    if not name:
        return logging.getLogger('sopel')

    parts = name.strip().split('.')
    if len(parts) > 1 or parts[0] in ['sopel', 'sopel_modules']:
        return logging.getLogger(name)

    # assume it's a plugin name, as intended by the original get_logger
    return tools.get_logger(name)



"""
find_updates.py - Sopel Update Check Plugin
This is separated from version.py, so that it can be easily overridden by
distribution packagers, and they can check their repositories rather than the
Sopel website.
Copyright 2014, Elsie Powell, embolalia.com
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

import json

import requests

from sopel import (
    __version__ as current_version,
    _version_info,
    plugin,
    tools,
    version_info,
)


wait_time = 24 * 60 * 60  # check once per day
max_failures = 4
version_url = 'https://sopel.chat/latest.json'
stable_message = (
    'A new Sopel version, {}, is available; I am running {}. Please update '
    'me. Full release notes at {}'
)
unstable_message = (
    'A new pre-release version, {}, is available; I am running {}. Please '
    'update me.{}'
)


@plugin.event(tools.events.RPL_LUSERCLIENT)
def startup_version_check(bot, trigger):
    if not bot.memory.get('update_startup_check_run', False):
        bot.memory['update_startup_check_run'] = True
        check_version(bot)


def _check_succeeded(bot):
    bot.memory['update_failures'] = 0


def _check_failed(bot):
    bot.memory['update_failures'] = 1 + bot.memory.get('update_failures', 0)


@plugin.interval(wait_time)
def check_version(bot):
    version = version_info
    r = None
    success = False

    try:
        r = requests.get(version_url, timeout=(5, 5))
    except requests.exceptions.RequestException:
        _check_failed(bot)
    else:
        success = True

    try:
        if r is not None:
            info = r.json()
    except json.JSONDecodeError:
        _check_failed(bot)
        success = False

    if not success:
        if bot.memory.get('update_failures', 0) <= max_failures:
            # not enough failures to worry; silently ignore this one
            return

        # too many failures to ignore; notify owner
        bot.say(
            "[update] I haven't been able to check for updates in a while. "
            "Please verify that {} is working and I can reach it."
            .format(version_url), bot.config.core.owner)
        bot.say(
            "[update] If this issue persists, please alert the Sopel dev team "
            "in #sopel on Libera Chat, or open a GitHub issue: "
            "https://github.com/sopel-irc/sopel/issues",
            bot.config.core.owner)
        return

    _check_succeeded(bot)

    if version.releaselevel == 'final':
        latest = info['version']
        notes = info['release_notes']
        message = stable_message
    else:
        latest = info['unstable']
        notes = info.get('unstable_notes', '')
        if notes:
            notes = ' Full release notes at ' + notes
        message = unstable_message
    latest_version = _version_info(latest)

    if version < latest_version:
        msg = message.format(latest, current_version, notes)
        bot.say('[update] ' + msg, bot.config.core.owner)



"""
lmgtfy.py - Sopel Let Me Google That For You Plugin
Copyright 2013, Dimitri Molenaars http://tyrope.nl/
Licensed under the Eiffel Forum License 2.

https://sopel.chat/
"""
from __future__ import annotations

from urllib.parse import urlencode

from sopel import plugin


@plugin.command('lmgtfy', 'lmgify', 'gify', 'gtfy')
@plugin.example('.lmgtfy sopel', 'https://lmgtfy.com/?q=sopel')
@plugin.example('.lmgtfy sopel bot', 'https://lmgtfy.com/?q=sopel+bot', user_help=True)
@plugin.example('.lmgtfy', 'https://www.google.com/', user_help=True)
def googleit(bot, trigger):
    """Let me just… Google that for you."""
    if not trigger.group(2):  # No input
        return bot.say('https://www.google.com/')
    qs = urlencode({
        'q': trigger.group(2),
    })
    bot.say('https://lmgtfy.com/?%s' % qs)



"""Constants for user privileges in channels."""
# Copyright 2023 dgw, technobabbl.es
#
# Licensed under the Eiffel Forum License 2.
from __future__ import annotations

import enum


__all__ = [
    'AccessLevel',
]


class AccessLevel(enum.IntFlag):
    """Enumeration of available user privilege levels.

    This class represents privileges as comparable, combinable flags. Lower
    privilege levels compare as *less than* (``<``) higher ones::

        >>> from sopel.privileges import AccessLevel
        >>> AccessLevel.VOICE < AccessLevel.HALFOP < AccessLevel.OP \\
        ... < AccessLevel.ADMIN < AccessLevel.OWNER
        True

    A user's privileges are represented as a combination of privilege levels::

        >>> priv = AccessLevel.VOICE | AccessLevel.OP

    This allows using comparators and bitwise operators to compare privileges.
    Here, ``priv`` contains both VOICE and OP privileges, but not HALFOP::

        >>> priv >= AccessLevel.OP
        True
        >>> bool(priv & AccessLevel.HALFOP)
        False

    .. important::

        Do not hard-code the value of a privilege level in your code; the values
        may change. Always reference or compare to the appropriate member of
        this class directly.

    """
    # values should behave as ints, but their string representations should
    # still look like Enum
    __str__ = enum.Enum.__str__

    VOICE = enum.auto()
    """Privilege level for the +v channel permission

    .. versionadded:: 4.1
    .. versionchanged:: 8.0

        Constant moved from :mod:`sopel.plugin` to
        :class:`sopel.privileges.AccessLevel`.

    """

    HALFOP = enum.auto()
    """Privilege level for the +h channel permission

    .. versionadded:: 4.1
    .. versionchanged:: 8.0

        Constant moved from :mod:`sopel.plugin` to
        :class:`sopel.privileges.AccessLevel`.

    .. important::

        Beware: This is one of the `nonstandard privilege levels`_.

    """

    OP = enum.auto()
    """Privilege level for the +o channel permission

    .. versionadded:: 4.1
    .. versionchanged:: 8.0

        Constant moved from :mod:`sopel.plugin` to
        :class:`sopel.privileges.AccessLevel`.

    """

    ADMIN = enum.auto()
    """Privilege level for the +a channel permission

    .. versionadded:: 4.1
    .. versionchanged:: 8.0

        Constant moved from :mod:`sopel.plugin` to
        :class:`sopel.privileges.AccessLevel`.

    .. important::

        Beware: This is one of the `nonstandard privilege levels`_.

    """

    OWNER = enum.auto()
    """Privilege level for the +q channel permission

    .. versionadded:: 4.1
    .. versionchanged:: 8.0

        Constant moved from :mod:`sopel.plugin` to
        :class:`sopel.privileges.AccessLevel`.

    .. important::

        Beware: This is one of the `nonstandard privilege levels`_.

    """

    OPER = enum.auto()
    """Privilege level for the +y/+Y channel permission

    Note: Except for these (non-standard) channel modes, Sopel does not monitor
    or store any user's OPER status.

    .. versionadded:: 7.0
    .. versionchanged:: 8.0

        Constant moved from :mod:`sopel.plugin` to
        :class:`sopel.privileges.AccessLevel`.

    .. important::

        Beware: This is one of the `nonstandard privilege levels`_.

    """



"""clock.py - Sopel Clock Plugin

Copyright 2008-9, Sean B. Palmer, inamidst.com
Copyright 2012, Elsie Powell, embolalia.com
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

from sopel import plugin
from sopel.tools.time import (
    format_time,
    get_channel_timezone,
    get_nick_timezone,
    get_timezone,
    validate_timezone
)

PLUGIN_OUTPUT_PREFIX = '[clock] '
ERROR_NO_TIMEZONE = (
    'What timezone do you want to use? '
    'Try one from https://sopel.chat/tz')
ERROR_NO_FORMAT = (
    'What format do you want me to use? '
    'Try using http://strftime.net to make one.')
ERROR_INVALID_TIMEZONE = (
    'Could not find timezone "%s". '
    'Try one from https://sopel.chat/tz')
ERROR_INVALID_FORMAT = (
    "That format doesn't work. Try using "
    "http://strftime.net to make one.")


@plugin.command('t', 'time')
@plugin.example('.t America/New_York', user_help=True)
@plugin.example('.t Exirel', user_help=True)
@plugin.example('.t #sopel', user_help=True)
@plugin.example('.t', user_help=True)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def f_time(bot, trigger):
    """Return the current time.

    The command takes an optional parameter: it will try to guess if it's a
    nick, a channel, or a timezone (in that order).

    If it's a known nick or channel but there is no configured timezone, then
    it will complain. If nothing can be found, it'll complain that the argument
    is not a valid timezone.

    The format used can be set with the ``settf`` or ``setctf`` commands.
    """
    argument = trigger.group(2)

    if not argument:
        # get default timezone from nick, or sender, or bot, or UTC
        zone = get_timezone(
            bot.db, bot.config, None, trigger.nick, trigger.sender)
    else:
        # guess if the argument is a nick, a channel, or a timezone
        zone = None
        argument = argument.strip()
        channel_or_nick = bot.make_identifier(argument)

        # first, try to get nick or channel's timezone
        help_prefix = bot.config.core.help_prefix
        if channel_or_nick.is_nick():
            zone = get_nick_timezone(bot.db, channel_or_nick)
            if zone is None and channel_or_nick in bot.users:
                # zone not found for a known nick: error case
                set_command = '%ssettz <zone>' % help_prefix
                if channel_or_nick != trigger.nick:
                    bot.reply(
                        'Could not find a timezone for this nick. '
                        '%s can set a timezone with `%s`'
                        % (argument, set_command))
                else:
                    bot.reply(
                        'Could not find a timezone for you. '
                        'You can set your timezone with `%s`'
                        % set_command)
                return
        else:
            zone = get_channel_timezone(bot.db, channel_or_nick)
            if zone is None and channel_or_nick in bot.channels:
                # zone not found for an existing channel: error case
                set_command = '%ssetctz <zone>' % help_prefix
                bot.reply(
                    'Could not find timezone for channel %s. '
                    'It can be set with `%s`. (requires OP privileges)'
                    % (argument, set_command))
                return

        # then, fallback on timezone detection
        if zone is None:
            # argument not found as nick or channel timezone
            try:
                zone = validate_timezone(argument)
            except ValueError:
                bot.reply(ERROR_INVALID_TIMEZONE % argument)
                return

    time = format_time(bot.db, bot.config, zone, trigger.nick, trigger.sender)
    bot.say(time)


@plugin.command('tz', 'timez')
@plugin.example('.tz America/New_York')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def f_time_zone(bot, trigger):
    """Return the current time in a timezone.

    Unlike the ``time`` command, it requires an argument, and that argument
    must be a valid timezone.
    """
    argument = trigger.group(2)
    if not argument:
        bot.reply(ERROR_NO_TIMEZONE)
        return

    zone = None
    argument = argument.strip()
    try:
        zone = validate_timezone(argument)
    except ValueError:
        bot.reply(ERROR_INVALID_TIMEZONE % argument)
        return

    time = format_time(bot.db, bot.config, zone, trigger.nick, trigger.sender)
    bot.say(time)


@plugin.command('settz', 'settimezone')
@plugin.example('.settz America/New_York')
def update_user(bot, trigger):
    """Set your preferred timezone.

    Most timezones will work, but it's best to use one from
    <https://sopel.chat/tz>.
    """
    argument = trigger.group(2)
    if not argument:
        bot.reply(ERROR_NO_TIMEZONE)
        return

    try:
        zone = validate_timezone(argument)
    except ValueError:
        bot.reply(ERROR_INVALID_TIMEZONE % argument)
        return

    bot.db.set_nick_value(trigger.nick, 'timezone', zone)

    if len(zone) < 4:
        bot.reply(
            'Okay, but you should use one from https://sopel.chat/tz '
            'if you use DST.')
    else:
        bot.reply('I now have you in the %s timezone.' % zone)


@plugin.command('unsettz', 'unsettimezone')
@plugin.example('.unsettz')
def unset_user_tz(bot, trigger):
    """Unset your preferred timezone."""
    bot.db.delete_nick_value(trigger.nick, 'timezone')
    bot.reply('Successfully unset timezone')


@plugin.command('gettz', 'gettimezone')
@plugin.example('.gettz Exirel', user_help=True)
@plugin.example('.gettz', user_help=True)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def get_user_tz(bot, trigger):
    """Gets a user's preferred time zone.

    It will show yours if no user specified.
    """
    nick = trigger.group(2)
    if not nick:
        nick = trigger.nick

    nick = nick.strip()
    zone = get_nick_timezone(bot.db, nick)

    if zone:
        bot.say('%s\'s time zone is %s.' % (nick, zone))
    else:
        bot.say('%s has not set their time zone' % nick)


@plugin.command('settimeformat', 'settf')
@plugin.example('.settf %Y-%m-%dT%T%z')
def update_user_format(bot, trigger):
    """Sets your preferred format for time.

    Uses the standard strftime format. You can use <http://strftime.net> or
    your favorite search engine to learn more.
    """
    tformat = trigger.group(2)
    if not tformat:
        bot.reply(ERROR_NO_FORMAT)
        return

    tz = get_timezone(bot.db, bot.config, None, trigger.nick, trigger.sender)

    # Get old format as back-up
    old_format = bot.db.get_nick_value(trigger.nick, 'time_format')

    # Save the new format in the database so we can test it.
    bot.db.set_nick_value(trigger.nick, 'time_format', tformat)

    try:
        timef = format_time(db=bot.db, zone=tz, nick=trigger.nick)
    except Exception:  # TODO: Be specific
        bot.reply(ERROR_INVALID_FORMAT)
        # New format doesn't work. Revert save in database.
        bot.db.set_nick_value(trigger.nick, 'time_format', old_format)
        return

    set_command = '%ssettz' % bot.settings.core.help_prefix
    bot.reply('Your time will now appear as %s. '
              '(If the timezone is wrong, you might try the %s command)'
              % (timef, set_command))


@plugin.command('unsettimeformat', 'unsettf')
@plugin.example('.unsettf')
def unset_user_format(bot, trigger):
    """Unsets your preferred format for time."""
    bot.db.delete_nick_value(trigger.nick, 'time_format')
    bot.reply('Successfully unset time format')


@plugin.command('gettimeformat', 'gettf')
@plugin.example('.gettf Exirel', user_help=True)
@plugin.example('.gettf', user_help=True)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def get_user_format(bot, trigger):
    """Gets a user's preferred time format.

    It will show yours if no user specified.
    """
    nick = trigger.group(2)
    if not nick:
        nick = trigger.nick

    nick = nick.strip()

    # Get old format as back-up
    format = bot.db.get_nick_value(nick, 'time_format')

    if format:
        bot.say("%s's time format: %s." % (nick, format))
    else:
        bot.say("%s hasn't set a custom time format" % nick)


@plugin.command('setchanneltz', 'setctz')
@plugin.example('.setctz America/New_York')
@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP, message='Changing the channel timezone requires OP privileges.')
def update_channel(bot, trigger):
    """Set the preferred timezone for the current channel."""
    argument = trigger.group(2)
    if not argument:
        bot.reply(ERROR_NO_TIMEZONE)
        return

    try:
        zone = validate_timezone(argument)
    except ValueError:
        bot.reply(ERROR_INVALID_TIMEZONE % argument)
        return

    channel = trigger.sender
    bot.db.set_channel_value(channel, 'timezone', zone)

    if len(zone) < 4:
        bot.reply(
            'Okay, but you should use one from https://sopel.chat/tz '
            'if you use DST.')
    else:
        bot.reply('I now have %s in the %s timezone.' % (channel, zone))


@plugin.command('unsetchanneltz', 'unsetctz')
@plugin.example('.unsetctz')
@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP, message='Changing the channel timezone requires OP privileges.')
def unset_channel(bot, trigger):
    """Unset the preferred timezone for the current channel."""
    channel = trigger.sender
    bot.db.delete_channel_value(channel, 'timezone')
    bot.reply('Successfully unset channel timezone')


@plugin.command('getchanneltz', 'getctz')
@plugin.example('.getctz #sopel', user_help=True)
@plugin.example('.getctz', user_help=True)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def get_channel_tz(bot, trigger):
    """Gets the channel's preferred timezone.

    It returns the current channel's if no channel name is given.
    """
    channel = trigger.group(2)
    if not channel:
        channel = trigger.sender

    channel = channel.strip()
    zone = get_channel_timezone(bot.db, channel)

    if zone:
        bot.say('%s\'s timezone: %s' % (channel, zone))
    else:
        bot.say('%s has no preferred timezone' % channel)


@plugin.command('setchanneltimeformat', 'setctf')
@plugin.example('.setctf %Y-%m-%dT%T%z')
@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def update_channel_format(bot, trigger):
    """Sets your preferred format for time.

    Uses the standard strftime format. You can use <http://strftime.net> or
    your favorite search engine to learn more.
    """
    tformat = trigger.group(2)
    if not tformat:
        bot.reply(ERROR_NO_FORMAT)
        return

    tz = get_timezone(bot.db, bot.config, None, None, trigger.sender)

    # Get old format as back-up
    old_format = bot.db.get_channel_value(trigger.sender, 'time_format')

    # Save the new format in the database so we can test it.
    bot.db.set_channel_value(trigger.sender, 'time_format', tformat)

    try:
        timef = format_time(db=bot.db, zone=tz, channel=trigger.sender)
    except Exception:  # TODO: Be specific
        bot.reply(ERROR_INVALID_FORMAT)
        # New format doesn't work. Revert save in database.
        bot.db.set_channel_value(trigger.sender, 'time_format', old_format)
        return
    bot.db.set_channel_value(trigger.sender, 'time_format', tformat)

    help_prefix = bot.settings.core.help_prefix
    set_command = '%ssettz' % help_prefix
    channel_command = '%schanneltz' % help_prefix
    bot.say("Times in this channel will now appear as %s "
            "unless a user has their own format set. (If the timezone"
            " is wrong, you might try the %s and %s "
            "commands)" % (timef, set_command, channel_command))


@plugin.command('unsetchanneltimeformat', 'unsetctf')
@plugin.example('.unsetctf')
@plugin.require_chanmsg
@plugin.require_privilege(plugin.OP, message='Changing the channel time format requires OP privileges.')
def unset_channel_format(bot, trigger):
    """Unset the preferred time format for the current channel."""
    channel = trigger.sender
    bot.db.delete_channel_value(channel, 'time_format')
    bot.reply('Successfully unset channel time format')


@plugin.command('getchanneltimeformat', 'getctf')
@plugin.example('.getctf #sopel', user_help=True)
@plugin.example('.getctf', user_help=True)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def get_channel_format(bot, trigger):
    """Gets the channel's preferred time format

    It returns the current channel's if no channel name is given.
    """
    channel = trigger.group(2)
    if not channel:
        channel = trigger.sender

    channel = channel.strip()

    tformat = bot.db.get_channel_value(channel, 'time_format')
    if tformat:
        bot.say('%s\'s time format: %s' % (channel, tformat))
    else:
        bot.say('%s has no preferred time format' % channel)



"""help.py - Obsolete Sopel Help Plugin

Install ``sopel-help`` with ``pip install sopel-help`` to get the official
help plugin for Sopel.
"""
from __future__ import annotations

import logging


LOGGER = logging.getLogger(__name__)


def setup(bot):
    LOGGER.warning(
        'Sopel\'s built-in help plugin is obsolete. '
        'Install sopel-help as the official help plugin for Sopel.\n'
        'You can install sopel-help with "pip install sopel-help".')



"""Thread-safe memory data-structures for Sopel.

Sopel uses lots of threads to manage rules and jobs and other features, and it
needs to store shared information safely. This class contains various memory
classes that are thread-safe, with some convenience features.
"""
from __future__ import annotations

from collections import defaultdict
import threading
from typing import Any, TYPE_CHECKING, Union

from typing_extensions import override

from .identifiers import Identifier, IdentifierFactory

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping
    from typing import Tuple

    MemoryConstructorInput = Union[
        Mapping[str, Any],
        Iterable[Tuple[str, Any]],
    ]


class _NO_DEFAULT:
    """Private class to help with overriding C methods like ``dict.pop()``.

    Some Python standard library features are implemented in pure C, and can
    have a ``null`` default value for certain parameters that is impossible to
    emulate at the Python layer. This class is our workaround for that.

    .. warning::

        Plugin authors **SHOULD NOT** use this class. It is not part of Sopel's
        public API.

    """


class SopelMemory(dict):
    """A simple thread-safe ``dict`` implementation.

    In order to prevent exceptions when iterating over the values and changing
    them at the same time from different threads, we use a blocking lock in
    ``__setitem__`` and ``__contains__``.

    .. note::

        Unlike the :class:`dict` on which they are based, ``SopelMemory`` and
        its derivative types do not accept key-value pairs as keyword arguments
        at construction time.

    .. versionadded:: 3.1
        As ``Willie.WillieMemory``
    .. versionchanged:: 4.0
        Moved to ``tools.WillieMemory``
    .. versionchanged:: 6.0
        Renamed from ``WillieMemory`` to ``SopelMemory``
    .. versionchanged:: 8.0
        Moved from ``tools`` to ``tools.memories``
    """
    def __init__(self, *args):
        dict.__init__(self, *args)
        self.lock = threading.Lock()

    def __setitem__(self, key, value):
        """Set a key equal to a value.

        The dict is locked for other writes while doing so.
        """
        self.lock.acquire()
        result = dict.__setitem__(self, key, value)
        self.lock.release()
        return result

    def __contains__(self, key):
        """Check if a key is in the dict.

        The dict is locked for writes while doing so.
        """
        self.lock.acquire()
        result = dict.__contains__(self, key)
        self.lock.release()
        return result


class SopelMemoryWithDefault(defaultdict):
    """Same as SopelMemory, but subclasses from collections.defaultdict.

    .. note::

        Unlike the :class:`~collections.defaultdict` on which it is based,
        ``SopelMemoryWithDefault`` does not accept key-value pairs as keyword
        arguments at construction time.

    .. versionadded:: 4.3
        As ``WillieMemoryWithDefault``
    .. versionchanged:: 6.0
        Renamed to ``SopelMemoryWithDefault``
    .. versionchanged:: 8.0
        Moved from ``tools`` to ``tools.memories``
    """
    def __init__(self, *args):
        defaultdict.__init__(self, *args)
        self.lock = threading.Lock()

    def __setitem__(self, key, value):
        """Set a key equal to a value.

        The dict is locked for other writes while doing so.
        """
        self.lock.acquire()
        result = defaultdict.__setitem__(self, key, value)
        self.lock.release()
        return result

    def __contains__(self, key):
        """Check if a key is in the dict.

        The dict is locked for writes while doing so.
        """
        self.lock.acquire()
        result = defaultdict.__contains__(self, key)
        self.lock.release()
        return result


class SopelIdentifierMemory(SopelMemory):
    """Special Sopel memory that stores ``Identifier`` as key.

    This is a convenient subclass of :class:`SopelMemory` that always casts its
    keys as instances of :class:`~.identifiers.Identifier`::

        >>> from sopel import tools
        >>> memory = tools.SopelIdentifierMemory()
        >>> memory['Exirel'] = 'king'
        >>> list(memory.items())
        [(Identifier('Exirel'), 'king')]
        >>> tools.Identifier('exirel') in memory
        True
        >>> 'exirel' in memory
        True

    As seen in the example above, it is possible to perform various operations
    with both ``Identifier`` and :class:`str` objects, taking advantage of the
    case-insensitive behavior of ``Identifier``.

    As it works with :class:`~.identifiers.Identifier`, it accepts an
    identifier factory. This factory usually comes from a
    :class:`bot instance<sopel.bot.Sopel>`, like in the example of a plugin
    setup function::

        def setup(bot):
            bot.memory['my_plugin_storage'] = SopelIdentifierMemory(
                identifier_factory=bot.make_identifier,
            )

    .. note::

        Internally, it will try to do ``key = self.make_identifier(key)``,
        which will raise an exception if it cannot instantiate the key
        properly::

            >>> memory[1] = 'error'
            AttributeError: 'int' object has no attribute 'translate'

    .. versionadded:: 7.1

    .. versionchanged:: 8.0

        Moved from ``tools`` to ``tools.memories``.

        The parameter ``identifier_factory`` has been added to properly
        transform ``str`` into :class:`~.identifiers.Identifier`. This factory
        is stored and accessible through :attr:`make_identifier`.

    """
    def __init__(
        self,
        *args: MemoryConstructorInput,
        identifier_factory: IdentifierFactory = Identifier,
    ) -> None:
        if len(args) > 1:
            raise TypeError(
                'SopelIdentifierMemory expected at most 1 argument, got {}'
                .format(len(args))
            )

        self.make_identifier = identifier_factory
        """A factory to transform keys into identifiers."""

        if len(args) == 1:
            super().__init__(self._convert_keys(args[0]))
        else:
            super().__init__()

    def _make_key(self, key: str | None) -> Identifier | None:
        if key is None:
            return None
        return self.make_identifier(key)

    def _convert_keys(
        self,
        data: MemoryConstructorInput,
    ) -> Iterable[tuple[Identifier, Any]]:
        """Ensure input keys are converted to ``Identifer``.

        :param data: the data passed to the memory at init or update
        :return: a generator of key-value pairs with the keys converted
                 to :class:`~.identifiers.Identifier`

        This private method takes input of a mapping or an iterable of key-value
        pairs and outputs a generator of key-value pairs ready for use in a new
        or updated :class:`self` instance. It is designed to work with any of the
        possible ways initial data can be passed to a :class:`dict`, except that
        ``kwargs`` must be passed to this method as a dictionary.
        """
        # figure out what to generate from
        if hasattr(data, 'items'):
            data = data.items()

        # return converted input data
        return ((self.make_identifier(k), v) for k, v in data)

    def __getitem__(self, key: str | None) -> Any:
        return super().__getitem__(self._make_key(key))

    def __contains__(self, key: Any) -> Any:
        return super().__contains__(self._make_key(key))

    def __setitem__(self, key: str | None, value: Any) -> None:
        super().__setitem__(self._make_key(key), value)

    def setdefault(self, key: str, default: Any = None) -> Any:
        return super().setdefault(self._make_key(key), default)

    def __delitem__(self, key: str) -> None:
        super().__delitem__(self._make_key(key))

    def copy(self):
        """Get a shallow copy of this ``SopelIdentifierMemory``.

        See :meth:`dict.copy`.
        """
        return type(self)(self, identifier_factory=self.make_identifier)

    def get(self, key: str, default: Any = _NO_DEFAULT) -> Any:
        """Get the value of ``key`` from this ``SopelIdentifierMemory``.

        Takes an optional ``default`` value, just like :meth:`dict.get`.
        """
        if default is _NO_DEFAULT:
            return super().get(self._make_key(key))
        return super().get(self._make_key(key), default)

    def pop(self, key: str, default: Any = _NO_DEFAULT) -> Any:
        """Pop the value of ``key`` from this ``SopelIdentifierMemory``.

        Takes an optional ``default`` value, just like :meth:`dict.pop`.
        """
        if default is _NO_DEFAULT:
            return super().pop(self._make_key(key))
        return super().pop(self._make_key(key), default)

    @override
    def update(self, maybe_mapping=tuple()):
        """Update this ``SopelIdentifierMemory`` with key-value pairs.

        See :meth:`dict.update`.
        """
        super().update(self._convert_keys(maybe_mapping))

    def __or__(self, other):
        if not isinstance(other, dict):
            return NotImplemented

        # self on the left, so other's keys overwrite
        new = self.copy()
        new.update(other)
        return new

    def __ror__(self, other):
        if not isinstance(other, dict):
            return NotImplemented

        # self on the right, so keep only new keys from other
        new = self.copy()
        new.update((k, v) for k, v in other.items() if k not in self)
        return new

    def __ior__(self, other):
        if not isinstance(other, dict):
            return NotImplemented
        self.update(other)
        return self

    def __eq__(self, other):
        if not isinstance(other, dict):
            return NotImplemented
        return super().__eq__(other)

    def __ne__(self, other):
        ret = self.__eq__(other)
        return ret if ret is NotImplemented else not ret
